from typing import Any, Dict, Literal, List

from ...modifiable_values import CombatActionValue, CostValue

from ...dice import Dice

from .base import (
    LimitedEffectSupportBase, RoundEffectSupportBase, SupportBase, 
    UsageWithRoundRestrictionSupportBase
)
from ...consts import (
    ELEMENT_DEFAULT_ORDER, CostLabels, DamageElementalType, DamageType, 
    DieColor, ElementType, ELEMENT_TO_DIE_COLOR, ElementalReactionType, 
    IconType, ObjectPositionType, PlayerActionLabels, SkillType
)
from ...struct import Cost, ObjectPosition
from ...action import (
    Actions, ChangeObjectUsageAction, ChargeAction, CreateDiceAction, 
    DrawCardAction, MoveObjectAction, RemoveDiceAction, RemoveObjectAction
)
from ...event import (
    ActionEndEventArguments, ChangeObjectUsageEventArguments, 
    ChooseCharactorEventArguments, MoveObjectEventArguments, 
    PlayerActionStartEventArguments, 
    ReceiveDamageEventArguments,
    RemoveObjectEventArguments, RoundEndEventArguments, 
    RoundPrepareEventArguments, SkillEndEventArguments, 
    SwitchCharactorEventArguments, UseCardEventArguments
)


class CompanionBase(SupportBase):
    cost_label: int = (CostLabels.CARD.value 
                       | CostLabels.COMPANION.value)


class RoundEffectCompanionBase(RoundEffectSupportBase):
    cost_label: int = (CostLabels.CARD.value 
                       | CostLabels.COMPANION.value)


class Paimon(CompanionBase):
    name: Literal['Paimon']
    desc: str = '''When Action Phase begins: Create Omni Element x2.'''
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 3)
    usage: int = 2
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateDiceAction | RemoveObjectAction]:
        """
        When in round prepare, increase usage. 
        If usage is 3, remove self, draw a card and create a dice.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        self.usage -= 1
        return [CreateDiceAction(
            player_idx = self.position.player_idx,
            number = 2,
            color = DieColor.OMNI,
        )] + self.check_should_remove()


class Katheryne(RoundEffectCompanionBase):
    name: Literal['Katheryne']
    desc: str = (
        'When you perform "Switch Character": The switch is considered a Fast '
        'Action instead of a Combat Action. (Once per Round)'
    )
    version: Literal['3.6'] = '3.6'
    cost: Cost = Cost(same_dice_number = 1)
    max_usage_per_round: int = 1

    def value_modifier_COMBAT_ACTION(
        self, value: CombatActionValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> CombatActionValue:
        """
        Copied from team status Leave It to Me!
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return value
        if self.usage <= 0:
            # no usage
            return value
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True,
        ):
            # not self switch charactor, do nothing
            return value
        if value.action_label & PlayerActionLabels.SWITCH.value == 0:
            # not switch charactor, do nothing
            return value
        if not value.do_combat_action:
            # already quick action, do nothing
            return value
        # self switch charactor, change to quick action
        value.do_combat_action = False
        assert mode == 'REAL'
        self.usage -= 1
        return value


class Timaeus(UsageWithRoundRestrictionSupportBase):
    name: Literal['Timaeus']
    desc: str = (
        'Comes with 2 Transmutation Materials when played. '
        'End Phase: Gain 1 Transmutation Material. '
        'When playing an Artifact Card: If possible, spend Transmutation '
        'Materials equal to the total cost of the Artifact and equip this '
        'Artifact for free. (Once per Round)'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 2)
    max_usage_one_round: int = 1
    decrease_target: int = CostLabels.ARTIFACT.value
    usage: int = 2
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def play(self, match: Any) -> List[Actions]:
        self.usage = 2
        return []

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        When in round end, increase usage.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        self.usage += 1
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['REAL', 'TEST']
    ) -> CostValue:
        """
        When in support, and self equip, and usage is enough, and has round
        usage, decrease cost.
        """
        decrease_number = value.cost.total_dice_cost
        if (
            self.position.area != ObjectPositionType.SUPPORT
            or self.position.player_idx != value.position.player_idx
            or value.cost.label & self.decrease_target == 0
            or decrease_number > self.usage
            or decrease_number == 0
            or self.usage_this_round <= 0
        ):
            # not in support, not self player, not target equip, not enough 
            # usage this round or total usage, do nothing
            return value
        # decrease all cost
        for _ in range(decrease_number):
            value.cost.decrease_cost(None)
        if mode == 'REAL':
            self.usage_this_round -= 1
            self.usage -= decrease_number
        return value


class Wagner(Timaeus):
    name: Literal['Wagner']
    desc: str = (
        'Comes with 2 Forging Billet when played. '
        'End Phase: Gain 1 Forging Billet. '
        'When playing a Weapon Card: If possible, spend Forging Billet '
        'equal to the total cost of the Weapon and equip this '
        'Weapon for free. (Once per Round)'
    )
    decrease_target: int = CostLabels.WEAPON.value


class ChefMao(RoundEffectCompanionBase, LimitedEffectSupportBase):
    name: Literal['Chef Mao']
    desc: str = (
        'After playing a Food Event Card: Create 1 random Elemental Die. '
        '(Once per Round) '
        'The first time the effect is triggered, draw 1 random Food Event '
        'Card from your deck.'
    )
    version: Literal['4.1'] = '4.1'
    cost: Cost = Cost(same_dice_number = 1)
    max_usage_per_round: int = 1
    limited_usage: int = 1

    def _limited_action(self, match: Any) -> List[DrawCardAction]:
        return [DrawCardAction(
            player_idx = self.position.player_idx,
            number = 1,
            whitelist_cost_labels = CostLabels.FOOD.value,
            draw_if_filtered_not_enough = False,
        )]

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Any
    ) -> List[Actions]:
        """
        When use food card, create one random elemental die and do limited
        action.
        """
        if self.position.area == ObjectPositionType.SUPPORT:
            # on support area, do effect
            if self.usage <= 0:
                # usage is 0, do nothing
                return []
            if (
                event.action.card_position.player_idx 
                != self.position.player_idx
            ):
                # not our charactor use card, do nothing
                return []
            if event.card.cost.label & CostLabels.FOOD.value == 0:
                # not food card, do nothing
                return []
            # our use food card, generate die and do limited action
            self.usage -= 1
            return [CreateDiceAction(
                player_idx = self.position.player_idx,
                number = 1,
                different = True  # use different to avoid generating OMNI
            )] + self.do_limited_action(match)
        # otherwise, do normal response
        return super().event_handler_USE_CARD(event, match)


class Tubby(RoundEffectCompanionBase):
    name: Literal['Tubby']
    desc: str = (
        'When playing a Location Support Card: Spend 2 less Elemental Dice. '
        '(Once per Round)'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 2)
    max_usage_per_round: int = 1
    decrease_target: int = CostLabels.LOCATION.value
    decrease_cost: int = 2

    def use(self, mode: Literal['REAL', 'TEST'], result: List[bool]) -> None:
        """
        When used, check whether need to decrease usage.
        """
        if True in result:
            if mode == 'REAL':
                self.usage -= 1

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['REAL', 'TEST']
    ) -> CostValue:
        """
        If in support and self use target card, decrease cost by 2.
        """
        if self.usage <= 0:
            # no usage
            return value
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player
            return value
        if value.cost.label & self.decrease_target == 0:
            # not decrease target
            return value
        # decrease
        result = [
            value.cost.decrease_cost(None) for _ in range(self.decrease_cost)
        ]
        self.use(mode, result)
        return value


class Timmie(CompanionBase):
    name: Literal['Timmie']
    desc: str = (
        'Triggers automatically once per Round: This card gains 1 Pigeon. '
        'When this card gains 3 Pigeons, discard this card, then draw 1 card '
        'and create Genius Invokation TCG Omni Dice Omni Element x1.'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost()
    usage: int = 0
    max_usage: int = 3
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def play(self, match: Any) -> List[ChangeObjectUsageAction]:
        """
        When played, first reset usage to 0, then increase usage.
        """
        self.usage = 0
        return [ChangeObjectUsageAction(
            object_position = self.position,
            change_type = 'DELTA',
            change_usage = 1
        )]

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        When in round prepare, increase usage. 
        If usage is 3, remove self, draw a card and create a dice.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        return [ChangeObjectUsageAction(
            object_position = self.position,
            change_type = 'DELTA',
            change_usage = 1
        )]

    def event_handler_CHANGE_OBJECT_USAGE(
        self, event: ChangeObjectUsageEventArguments, match: Any
    ) -> List[RemoveObjectAction | DrawCardAction | CreateDiceAction]:
        """
        when self usage changed to 3, remove self, draw a card and create a 
        dice.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        if event.action.object_position.id != self.id:
            # not self
            return []
        ret: List[RemoveObjectAction | DrawCardAction | CreateDiceAction] = []
        if self.usage == 3:
            ret += [
                RemoveObjectAction(
                    object_position = self.position,
                ),
                DrawCardAction(
                    player_idx = self.position.player_idx,
                    number = 1,
                    draw_if_filtered_not_enough = True,
                ),
                CreateDiceAction(
                    player_idx = self.position.player_idx,
                    number = 1,
                    color = DieColor.OMNI,
                ),
            ]
        return ret


class Liben(CompanionBase):
    name: Literal['Liben']
    desc: str = (
        'End Phase: Collect your unused Elemental Dice (Max 1 of each '
        'Elemental Type). '
        'When Action Phase begins: If this card has collected 3 Elemental '
        'Dice, draw 2 cards and create Omni Element x2, then discard this '
        'card.'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost()
    usage: int = 0
    max_usage: int = 3
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def play(self, match: Any) -> List[Actions]:
        self.usage = 0
        return []

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[RemoveDiceAction]:
        """
        Collect different color dice in round end.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        collect_order: List[DieColor] = []
        dice: Dice = match.player_tables[self.position.player_idx].dice
        colors = dice.colors
        # collect die in element default order
        for element in ELEMENT_DEFAULT_ORDER:
            color = ELEMENT_TO_DIE_COLOR[element]
            if color in colors:
                collect_order.append(color)
        # collect OMNI at last
        for color in colors:
            if color == DieColor.OMNI:
                collect_order.append(color)
        # only collect dice if usage is not full
        collect_order = collect_order[: self.max_usage - self.usage]
        self.usage += len(collect_order)
        return [RemoveDiceAction(
            player_idx = self.position.player_idx,
            dice_idxs = dice.colors_to_idx(collect_order)
        )]

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateDiceAction | DrawCardAction | RemoveObjectAction]:
        """
        If this card has collected 3 Elemental Dice, draw 2 cards and create 
        Omni Element x2, then discard this card.'
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        if self.usage != self.max_usage:
            # not enough dice collected
            return []
        return [ 
            CreateDiceAction(
                player_idx = self.position.player_idx,
                number = 2,
                color = DieColor.OMNI,
            ),
            DrawCardAction(
                player_idx = self.position.player_idx,
                number = 2,
                draw_if_filtered_not_enough = True
            ),
            RemoveObjectAction(
                object_position = self.position,
            ),
        ]


class ChangTheNinth(CompanionBase):
    name: Literal['Chang the Ninth']
    desc: str = (
        'When either side uses a Skill: If Physical DMG or Piercing DMG was '
        'dealt, or an Elemental Reaction was triggered, this card gains 1 '
        'Inspiration. When this card gains 3 Inspiration, discard this card, '
        'then draw 2 cards.'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost()
    usage: int = 0
    max_usage: int = 3
    inspiration_got: bool = False
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def event_handler_PLAYER_ACTION_START(
        self, event: PlayerActionStartEventArguments, match: Any
    ) -> List[Actions]:
        """
        When anyone start action, reset inspiration got.
        """
        self.inspiration_got = False
        return []

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        If physical or piercing or elemental reaction, gain inspiration.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        if event.final_damage.damage_type == DamageType.HEAL:
            # heal, do nothing
            return []
        if (
            event.final_damage.damage_elemental_type not in [
                DamageElementalType.PHYSICAL, DamageElementalType.PIERCING]
            and event.final_damage.element_reaction 
            == ElementalReactionType.NONE
        ):
            # not physical piercing and not elemental reaction, do nothing
            return []
        # get inspiration
        self.inspiration_got = True
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[DrawCardAction | RemoveObjectAction]:
        """
        If inspiration got, increase usage. If usage receives max_usage, 
        remove self and draw two cards.
        """
        if self.inspiration_got:
            self.usage += 1
        self.inspiration_got = False
        if self.usage == self.max_usage:
            # remove self and draw two cards
            return [
                RemoveObjectAction(
                    object_position = self.position,
                ),
                DrawCardAction(
                    player_idx = self.position.player_idx,
                    number = 2,
                    draw_if_filtered_not_enough = True
                ),
            ]
        return []


class Ellin(RoundEffectCompanionBase):
    name: Literal['Ellin']
    desc: str = (
        'When you use a Skill that has already been used in this Round: Spend '
        '1 less Elemental Die. (Once per Round)'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 2)
    usage: int = 0
    max_usage_per_round: int = 1
    recorded_skill_ids: Dict[int, None] = {}

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        self.recorded_skill_ids.clear()
        return super().event_handler_ROUND_PREPARE(event, match)

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        Record skill id when skill end.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        if event.action.position.player_idx != self.position.player_idx:
            # not self player
            return []
        self.recorded_skill_ids[event.action.position.id] = None
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['REAL', 'TEST']
    ) -> CostValue:
        """
        When a skill id is recorded, and has usage, decrease cost.
        """
        if (
            self.position.area != ObjectPositionType.SUPPORT
            or self.usage <= 0
            or value.position.id not in self.recorded_skill_ids
            or value.position.player_idx != self.position.player_idx
            or value.cost.label & (
                CostLabels.NORMAL_ATTACK.value
                | CostLabels.ELEMENTAL_SKILL.value
                | CostLabels.ELEMENTAL_BURST.value
            ) == 0
        ):
            # no usage, not self player, not recorded skill id or not skill
            return value
        # decrease
        if value.cost.decrease_cost(value.cost.elemental_dice_color):
            if mode == 'REAL':
                self.usage -= 1
        return value


class IronTongueTian(CompanionBase):
    name: Literal['Iron Tongue Tian']
    desc: str = (
        'End Phase: One of your characters without maximum Energy gains 1 '
        'Energy. (Active Character prioritized)'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(any_dice_number = 2)
    usage: int = 2
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[ChargeAction | RemoveObjectAction]:
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        table = match.player_tables[self.position.player_idx]
        charactors = [table.get_active_charactor()]
        for c in match.player_tables[self.position.player_idx].charactors:
            if c.position.id != charactors[0].position.id:
                charactors.append(c)
        for c in charactors:
            if c.is_alive and c.charge < c.max_charge:
                assert self.usage > 0
                self.usage -= 1
                return [ChargeAction(
                    player_idx = c.position.player_idx,
                    charactor_idx = c.position.charactor_idx,
                    charge = 1,
                )] + self.check_should_remove()
        return []


class LiuSu(CompanionBase, UsageWithRoundRestrictionSupportBase):
    name: Literal['Liu Su']
    desc: str = (
        'After you switch characters: If the character you switched to does '
        'not have Energy, they will gain 1 Energy. (Once per Round)'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 1)
    usage: int = 2
    max_usage_one_round: int = 1
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def charge(self, charactor: Any) -> List[ChargeAction 
                                             | RemoveObjectAction]:
        """
        If this charactor is our charactor and has no energy, charge it.
        """
        if (
            self.position.area == ObjectPositionType.SUPPORT
            and charactor.position.player_idx == self.position.player_idx
            and charactor.charge == 0 
            and self.has_usage()
        ):
            # in support, and our charactor, and no energy
            self.use()
            return [ChargeAction(
                player_idx = charactor.position.player_idx,
                charactor_idx = charactor.position.charactor_idx,
                charge = 1,
            )] + self.check_should_remove()
        return []

    def event_handler_SWITCH_CHARACTOR(
        self, event: SwitchCharactorEventArguments, match: Any
    ) -> List[ChargeAction | RemoveObjectAction]:
        charactor = match.player_tables[event.action.player_idx].charactors[
            event.action.charactor_idx]
        return self.charge(charactor)

    def event_handler_CHOOSE_CHARACTOR(
        self, event: ChooseCharactorEventArguments, match: Any
    ) -> List[ChargeAction | RemoveObjectAction]:
        charactor = match.player_tables[event.action.player_idx].charactors[
            event.action.charactor_idx]
        return self.charge(charactor)


class Hanachirusato(CompanionBase):
    name: Literal['Hanachirusato']
    desc: str = (
        'When a Summon disappears: This card gains 1 Cleansing Ritual '
        'Progress. (Max 3). When you play a Weapon or Artifact Card: If you '
        'already have 3 Cleansing Ritual Progress, discard this card and '
        'cause the card you play to cost 2 less Elemental Dice.'
    )
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost()
    usage: int = 0
    max_usage: int = 3
    decrease_target: int = CostLabels.WEAPON.value | CostLabels.ARTIFACT.value
    decrease_number: int = 2
    effect_triggered: bool = False
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def event_handler_REMOVE_OBJECT(
        self, event: RemoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        """
        When any sommon is removed, add one usage
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        if event.action.object_position.area != ObjectPositionType.SUMMON:
            # not summon, do nothing
            return []
        self.usage = min(self.usage + 1, self.max_usage)
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['REAL', 'TEST']
    ) -> CostValue:
        """
        When in support, and self use target card, decrease cost by 2.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return value
        if self.usage != self.max_usage:
            # not enough usage
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player
            return value
        if value.cost.label & self.decrease_target == 0:
            # not decrease target
            return value
        # decrease
        result = [
            value.cost.decrease_cost(None),
            value.cost.decrease_cost(None),
        ]
        if True in result:
            if mode == 'REAL':
                self.effect_triggered = True
        return value

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        if self.position.area == ObjectPositionType.SUPPORT:
            # in support, check if effect triggered. If so, remove self.
            if self.effect_triggered:
                return [RemoveObjectAction(
                    object_position = self.position,
                )]
        return super().event_handler_MOVE_OBJECT(event, match)


class KidKujirai(CompanionBase):
    name: Literal['Kid Kujirai']
    desc: str = (
        "When the Action Phase begins: Create 1 Omni Element. Then if your "
        "opponent's Support Zone is not full, transfer this card to your "
        "opponent's Support Zone."
    )
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost()
    usage: int = 0
    icon_type: Literal[IconType.NONE] = IconType.NONE

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateDiceAction | MoveObjectAction | RemoveObjectAction]:
        """
        Create one OMNI for this side; and if other side has position, move
        to opposite; otherwise remove self.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        ret: List[CreateDiceAction | MoveObjectAction 
                  | RemoveObjectAction] = []
        ret.append(CreateDiceAction(
            player_idx = self.position.player_idx,
            number = 1,
            color = DieColor.OMNI,
        ))
        opposite = match.player_tables[1 - self.position.player_idx].supports
        max_number = match.config.max_support_number
        if len(opposite) >= max_number:
            # opposite side support is full, remove self
            ret.append(RemoveObjectAction(
                object_position = self.position,
            ))
        else:
            # opposite side support is not full, move to opposite
            ret.append(MoveObjectAction(
                object_position = self.position,
                target_position = ObjectPosition(
                    player_idx = 1 - self.position.player_idx,
                    area = ObjectPositionType.SUPPORT,
                    id = self.position.id
                ),
            ))
        return ret


class Xudong(Tubby):
    name: Literal['Xudong']
    desc: str = (
        'When playing a Food Event Card: Spend 2 less Elemental Dice. '
        '(Once per Round)'
    )
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(any_dice_number = 2)
    decrease_target: int = CostLabels.FOOD.value
    decrease_cost: int = 2


class Dunyarzad(Tubby, LimitedEffectSupportBase):
    name: Literal['Dunyarzad']
    desc: str = (
        'When playing a Companion Support Card: Spend 1 less Elemental Dice. '
        '(Once per Round) '
        'The first time the effect is triggered, draw 1 random Companion '
        'Support Card from your deck.'
    )
    version: Literal['4.1'] = '4.1'
    cost: Cost = Cost(same_dice_number = 1)
    decrease_target: int = CostLabels.COMPANION.value
    decrease_cost: int = 1

    limited_usage: int = 1
    triggered: bool = False

    def _limited_action(self, match: Any) -> List[DrawCardAction]:
        return [DrawCardAction(
            player_idx = self.position.player_idx,
            number = 1,
            whitelist_cost_labels = self.decrease_target,
            draw_if_filtered_not_enough = False,
        )]

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Any
    ) -> List[Actions]:
        """
        Do limited action to draw a support card if needed
        """
        if self.position.area == ObjectPositionType.SUPPORT:
            # on support area, do effect
            if (
                event.action.card_position.player_idx 
                != self.position.player_idx
            ):
                # not our charactor use card, do nothing
                return []
            if event.card.cost.label & CostLabels.COMPANION.value == 0:
                # not companion card, do nothing
                return []
            # our use companion card, do limited action
            return self.do_limited_action(match)
        # otherwise, do normal response
        return super().event_handler_USE_CARD(event, match)


class Rana(RoundEffectCompanionBase):
    name: Literal['Rana']
    desc: str = (
        'After your character uses an Elemental Skill: '
        'Create 1 Elemental Die of the same Type as your next off-field '
        'character. (Once per Round)'
    )
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 2)
    max_usage_per_round: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        if it is in support are, and self player used a elemental skill,
        and have usage, and have next charactor, generate a die with color 
        of next charactor.
        """
        if (
            self.position.area == ObjectPositionType.SUPPORT
            and event.action.position.player_idx == self.position.player_idx 
            and event.action.skill_type == SkillType.ELEMENTAL_SKILL
            and self.usage > 0
        ):
            table = match.player_tables[self.position.player_idx]
            next_idx = table.next_charactor_idx()
            if next_idx is not None:
                self.usage -= 1
                ele_type: ElementType = table.charactors[next_idx].element
                die_color = ELEMENT_TO_DIE_COLOR[ele_type]
                return [CreateDiceAction(
                    player_idx = self.position.player_idx,
                    number = 1,
                    color = die_color,
                )]
        return []


class MasterZhang(RoundEffectCompanionBase):
    name: Literal['Master Zhang']
    desc: str = (
        'When playing a Weapon card: Spend 1 less Elemental Die. On top of '
        'that, for each of your characters already equipped with a Weapon on '
        'the field, you spend 1 less Elemental Die. (Once per Round.)'
    )
    version: Literal['3.8'] = '3.8'
    cost: Cost = Cost(same_dice_number = 1)
    max_usage_per_round: int = 1
    card_cost_label: int = CostLabels.WEAPON.value

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['REAL', 'TEST']
    ) -> CostValue:
        """
        If has usage, and self use weapon card, decrease.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return value
        if self.usage <= 0:
            # no usage
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player
            return value
        if value.cost.label & self.card_cost_label == 0:
            # not corresponding card
            return value
        # decrease
        decrease_number = 1
        table = match.player_tables[self.position.player_idx]
        for charactor in table.charactors:
            if charactor.weapon is not None:
                decrease_number += 1
        result = []
        for _ in range(decrease_number):
            result.append(value.cost.decrease_cost(None))
        if True in result:
            if mode == 'REAL':
                self.usage -= 1
        return value


class Setaria(CompanionBase):
    name: Literal['Setaria']
    desc: str = (
        'After you perform any action, if you have 0 cards in your hand: '
        'Draw 1 card.'
    )
    version: Literal['4.0'] = '4.0'
    cost: Cost = Cost(same_dice_number = 1)
    usage: int = 3
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ACTION_END(
        self, event: ActionEndEventArguments, match: Any
    ) -> List[DrawCardAction | RemoveObjectAction]:
        """
        If self action end and hand is empty, draw card.
        """
        if event.action.position.player_idx != self.position.player_idx:
            # not self player
            return []
        if len(match.player_tables[self.position.player_idx].hands) != 0:
            # not empty hand
            return []
        # draw a card
        self.usage -= 1
        return [DrawCardAction(
            player_idx = self.position.player_idx,
            number = 1,
            draw_if_filtered_not_enough = True,
        )] + self.check_should_remove()


class YayoiNanatsuki(MasterZhang):
    name: Literal['Yayoi Nanatsuki']
    desc: str = (
        'When playing an Artifact card: Spend 1 less Elemental Die. On top of '
        'that, for each of your characters already equipped with an artifact '
        'on the field, you spend 1 less Elemental Die. (Once per Round.)'
    )
    version: Literal['4.1'] = '4.1'
    card_cost_label: int = CostLabels.ARTIFACT.value


Companions = (
    Paimon | Katheryne | Timaeus | Wagner | ChefMao | Tubby | Timmie | Liben 
    | ChangTheNinth | Ellin | IronTongueTian | LiuSu | Hanachirusato 
    | KidKujirai | Xudong | Dunyarzad | Rana | MasterZhang | Setaria
    | YayoiNanatsuki
)



from typing import Any, List, Literal

from ...action import Actions

from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Skills


class GuhuaStyle(PhysicalNormalAttackBase):
    name: Literal['Guhua Style'] = 'Guhua Style'
    cost: Cost = PhysicalNormalAttackBase.get_cost(ElementType.HYDRO)


class FatalRainscreen(ElementalSkillBase):
    name: Literal['Fatal Rainscreen'] = 'Fatal Rainscreen'
    desc: str = (
        'Deals 2 Hydro DMG, grants this character Hydro Application, '
        'creates 1 Rain Sword.'
    )
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack, application and create object
        """
        ret = super().get_actions(match)
        ret += [
            self.element_application_self(match, DamageElementalType.HYDRO)
        ]
        if self.is_talent_equipped(match):
            ret += [self.create_team_status('Rain Sword', {
                'usage': 3,
                'max_usage': 3
            })]
        else:
            ret += [self.create_team_status('Rain Sword')]
        return ret


class Raincutter(ElementalBurstBase):
    name: Literal['Raincutter'] = 'Raincutter'
    desc: str = (
        'Deals 2 Hydro DMG, grants this character Hydro Application, creates '
        '1 Rainbow Bladework.'
    )
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack, application and create object
        """
        ret = super().get_actions(match)
        ret += [
            self.element_application_self(match, DamageElementalType.HYDRO)
        ]
        return ret + [
            self.create_team_status('Rainbow Bladework')
        ]


# Talents


class TheScentRemained(SkillTalent):
    name: Literal['The Scent Remained']
    desc: str = (
        'Combat Action: When your active character is Xingqiu, '
        'equip this card. After Xingqiu equips this card, immediately use '
        'Fatal Rainscreen once. When your Xingqiu, who has this card '
        'equipped, creates a Rain Sword, its starting Usage(s)+1.'
    )
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Xingqiu'] = 'Xingqiu'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 4
    )
    skill: FatalRainscreen = FatalRainscreen()


# charactor base


class Xingqiu(CharactorBase):
    name: Literal['Xingqiu']
    version: Literal['4.1'] = '4.1'
    desc: str = '''"Juvenile Galant" Xingqiu'''
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[GuhuaStyle | FatalRainscreen | Raincutter] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.SWORD
    talent: TheScentRemained | None = None

    def _init_skills(self) -> None:
        self.skills = [
            GuhuaStyle(),
            FatalRainscreen(),
            Raincutter(),
        ]



from typing import Literal


from ...consts import IconType

from .base import RoundCharactorStatus
from .cryo_charactors import Grimheart as G_3_8
from .foods import MintyMeatRolls as MMR_3_4
from .hydro_charactors import RangedStance as RS_4_1, MeleeStance as MS_4_1
from .hydro_charactors import Riptide as Riptide_4_1


class Grimheart_3_5(G_3_8):
    version: Literal['3.5']
    damage: int = 2


class MintyMeatRolls_3_3(MMR_3_4):
    desc: str = (
        "During this Round, the target character's Normal Attacks cost "
        "less 1 Unaligned Element."
    )
    version: Literal['3.3']
    decrease_usage: int = 999


class Riptide_3_7(Riptide_4_1, RoundCharactorStatus):
    """
    This status will not deal damages directly, and defeat-regenerate is 
    handled by system handler, which is created by Tartaglia'is passive skill
    when game starts.
    """
    name: Literal['Riptide'] = 'Riptide'
    desc: str = (
        'When the character to which this is attached is defeated: Apply '
        'Riptide to active character. '
        'When Tartaglia is in Melee Stance, he will deal additional DMG when '
        'attacking the character to which this is attached.'
    )
    version: Literal['3.7']
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS


class RangedStance_3_7(RS_4_1):
    version: Literal['3.7']


class MeleeStance_3_7(MS_4_1):
    version: Literal['3.7']


OldVersionCharactorStatus = (
    Riptide_3_7 | RangedStance_3_7 | MeleeStance_3_7

    | Grimheart_3_5 

    | MintyMeatRolls_3_3
)



from typing import Any, List, Literal

from ....modifiable_values import DamageIncreaseValue

from ....event import SkillEndEventArguments

from ....consts import ObjectPositionType, SkillType

from ....action import ChargeAction

from ....struct import Cost
from .base import ArtifactBase, RoundEffectArtifactBase


class OrnateKabuto(ArtifactBase):
    name: Literal['Ornate Kabuto'] = 'Ornate Kabuto'
    desc: str = (
        'After another character of yours uses an Elemental Burst: The '
        'character to which this is attached gains 1 Energy.'
    )
    version: Literal['4.0'] = '4.0'
    cost: Cost = Cost(same_dice_number = 1)
    usage: int = 0

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChargeAction]:
        if not (
            self.position.area == ObjectPositionType.CHARACTOR
            and event.action.position.player_idx == self.position.player_idx 
            and event.action.position.charactor_idx 
            != self.position.charactor_idx
            and event.action.skill_type == SkillType.ELEMENTAL_BURST
        ):
            # not equipped or not our other use burst
            return []
        # charge self by 1
        return [ChargeAction(
            player_idx = self.position.player_idx,
            charactor_idx = self.position.charactor_idx,
            charge = 1
        )]


class EmblemOfSeveredFate(OrnateKabuto, RoundEffectArtifactBase):
    name: Literal['Emblem of Severed Fate'] = 'Emblem of Severed Fate'
    desc: str = (
        'After a character uses an Elemental Burst: The character to which '
        'this is attached gains 1 Energy. The DMG dealt by the '
        "character's Elemental Bursts is increased by 2. (Once per Round)"
    )
    version: Literal['4.1'] = '4.1'
    cost: Cost = Cost(same_dice_number = 2)
    max_usage_per_round: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return value
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.ELEMENTAL_BURST
        ):
            # not self use elemental burst
            return value
        if self.usage <= 0:
            # no usage left
            return value
        # modify damage
        assert mode == 'REAL'
        self.usage -= 1
        value.damage += 2
        return value


EmblemOfSeveredFateArtifacts = OrnateKabuto | EmblemOfSeveredFate



from typing import Any, List, Literal

from ...event import (
    ChooseCharactorEventArguments, RoundPrepareEventArguments, 
    SkillEndEventArguments, SwitchCharactorEventArguments
)

from ...struct import Cost, ObjectPosition

from ...action import Actions, MakeDamageAction

from ...consts import (
    DamageElementalType, DamageType, IconType, SkillType, WeaponType
)

from ...modifiable_values import (
    DamageElementEnhanceValue, DamageIncreaseValue, DamageMultiplyValue, 
    DamageValue
)
from .base import (
    DefendTeamStatus, ElementalInfusionTeamStatus, ExtraAttackTeamStatus, 
    RoundTeamStatus, UsageTeamStatus
)


class IllusoryBubble(UsageTeamStatus):
    """
    Team status generated by Mona.
    """
    name: Literal['Illusory Bubble'] = 'Illusory Bubble'
    desc: str = (
        'When dealing Skill DMG: Remove this status and double the DMG dealt '
        'for this instance.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def value_modifier_DAMAGE_MULTIPLY(
        self, value: DamageMultiplyValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageMultiplyValue:
        """
        Double damage when skill damage made.
        """
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not this charactor use skill, not modify
            return value
        if self.usage > 0:
            value.damage *= 2
            assert mode == 'REAL'
            self.usage -= 1
        return value


class RainbowBladework(UsageTeamStatus, ExtraAttackTeamStatus):
    name: Literal['Rainbow Bladework'] = 'Rainbow Bladework'
    desc: str = 'After your character uses a Normal Attack: Deal 1 Hydro DMG.'
    version: Literal['3.6'] = '3.6'
    usage: int = 3
    max_usage: int = 3
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    trigger_skill_type: SkillType | None = SkillType.NORMAL_ATTACK
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    decrease_usage: bool = True


class RainSword(DefendTeamStatus):
    name: Literal['Rain Sword'] = 'Rain Sword'
    desc: str = (
        'When your active character receives at least 3 DMG: '
        'Decrease DMG taken by 1.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    min_damage_to_trigger: int = 3
    max_in_one_time: int = 1


class PrayerOfTheCrimsonCrown(ElementalInfusionTeamStatus, 
                              RoundTeamStatus,
                              ExtraAttackTeamStatus):
    name: Literal[
        'Prayer of the Crimson Crown'] = 'Prayer of the Crimson Crown'
    desc: str = (
        "Your characters' Normal Attacks deal +1 DMG. "
        "Your Sword, Claymore, and Polearm-wielding characters' Physical DMG "
        "is converted to Hydro DMG. After you switch characters, you deal 1 "
        "Hydro DMG. (Once per Round). "
    )
    version: Literal['3.8'] = '3.8'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    switch_damage_usage: int = 1
    switch_damage_max_usage: int = 1
    extra_damage_usage: int = 1
    extra_damage_max_usage: int = 1

    # enhance args
    infused_elemental_type: DamageElementalType = DamageElementalType.HYDRO

    # extra attack args
    trigger_skill_type: SkillType | None = SkillType.NORMAL_ATTACK
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    decrease_usage: bool = False

    # switch charactor attack related

    def _attack(
        self, match: Any, usage_type: Literal['SWITCH', 'EXTRA'], 
        damage: int = 1
    ) -> List[MakeDamageAction]:
        """
        attack enemy active charactor
        """
        if usage_type == 'SWITCH':
            if self.switch_damage_usage <= 0:
                # no usage
                return []
            self.switch_damage_usage -= 1
        else:
            assert usage_type == 'EXTRA'
            if self.extra_damage_usage <= 0:
                # no usage
                return []
            self.extra_damage_usage -= 1
        active_charactor = match.player_tables[
            1 - self.position.player_idx].get_active_charactor()
        return [MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = 1 - self.position.player_idx,
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.DAMAGE,
                    target_position = active_charactor.position,
                    damage = damage,
                    damage_elemental_type = DamageElementalType.HYDRO,
                    cost = Cost()
                )
            ]
        )]

    def event_handler_SWITCH_CHARACTOR(
        self, event: SwitchCharactorEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If self switch charactor, perform attack
        """
        if event.action.player_idx != self.position.player_idx:
            # not self switch charactor
            return []
        return self._attack(match, 'SWITCH')

    def event_handler_CHOOSE_CHARACTOR(
        self, event: ChooseCharactorEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If self choose charactor (when ally defeated), perform attack
        """
        if event.action.player_idx != self.position.player_idx:
            # not self switch charactor
            return []
        return self._attack(match, 'SWITCH')

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        reset usage
        """
        self.switch_damage_usage = self.switch_damage_max_usage
        self.extra_damage_usage = self.extra_damage_max_usage
        return super().event_handler_ROUND_PREPARE(event, match)

    # renew

    def renew(self, new_status: 'PrayerOfTheCrimsonCrown') -> None:
        self.switch_damage_max_usage = max(self.switch_damage_max_usage,
                                           new_status.switch_damage_max_usage)
        self.switch_damage_usage = max(self.switch_damage_usage,
                                       new_status.switch_damage_usage)
        self.extra_damage_max_usage = max(self.switch_damage_max_usage,
                                          new_status.extra_damage_max_usage)
        self.extra_damage_usage = max(self.switch_damage_usage,
                                      new_status.extra_damage_usage)
        return super().renew(new_status)

    # enhance and add damage

    def charactor_is_using_right_weapon(
        self, match: Any, position: ObjectPosition
    ) -> bool:
        """
        If position charactor using effected three weapon, return True.
        """
        charactor = match.player_tables[position.player_idx].charactors[
            position.charactor_idx]
        return charactor.weapon_type in [
            WeaponType.SWORD, WeaponType.CLAYMORE, WeaponType.POLEARM]

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self, value: DamageElementEnhanceValue, match: Any, 
        mode: Literal['TEST', 'REAL']
    ) -> DamageElementEnhanceValue:
        """
        If weapon type not right, do not enhance.
        """
        if not self.charactor_is_using_right_weapon(match, value.position):
            return value
        return super().value_modifier_DAMAGE_ELEMENT_ENHANCE(
            value, match, mode)

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        +1 our normal attack DMG.
        """
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not corresponding charactor, do nothing
            return value
        # add damage
        assert mode == 'REAL'
        value.damage += 1
        return value

    # extra attack

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If candace have talent, and our charactor use normal attack, then
        1 hydro damage.
        """
        if event.action.position.player_idx != self.position.player_idx:
            # not our player
            return []
        if event.action.skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack
            return []
        charactors = match.player_tables[self.position.player_idx].charactors
        candace_talent = False
        for charactor in charactors:
            if charactor.name == 'Candace' and charactor.talent is not None:
                candace_talent = True
                break
        if not candace_talent:
            # no candace with talent
            return []
        # make damage
        return self._attack(match, 'EXTRA')


HydroCharactorTeamStatus = (
    IllusoryBubble | RainbowBladework | RainSword | PrayerOfTheCrimsonCrown
)



from .kaeya import Kaeya, ColdBloodedStrike
from .shenhe import Shenhe, MysticalAbandon, TalismanSpirit
from .eula import Eula, LightfallSword, WellspringOfWarLust
from .chongyun import Chongyun, SteadyBreathing
from .ganyu import Ganyu, SacredCryoPearl, UndividedHeart
from .qiqi import Qiqi, RiteOfResurrection, HeraldOfFrost
from .kamisato_ayaka import (
    KamisatoAyaka, FrostflakeSekiNoTo, KantenSenmyouBlessing
)
from .diona import Diona, DrunkenMist, ShakenNotPurred
from .fatui_cryo_cicin_mage import (
    FatuiCryoCicinMage, CryoCicins, CicinsColdGlare
)


CryoCharactors = (
    Ganyu | Diona | Kaeya | Chongyun | KamisatoAyaka | Eula | Shenhe | Qiqi
    # finally monsters
    | FatuiCryoCicinMage
)
SummonsOfCryoCharactors = (
    SacredCryoPearl | DrunkenMist | FrostflakeSekiNoTo | LightfallSword 
    | TalismanSpirit | HeraldOfFrost
    # finally monsters
    | CryoCicins
)
CryoCharactorTalents = (
    UndividedHeart | ShakenNotPurred | ColdBloodedStrike | SteadyBreathing 
    | KantenSenmyouBlessing | WellspringOfWarLust | MysticalAbandon
    | RiteOfResurrection
    # finally monsters
    | CicinsColdGlare
)



from typing import List, Literal
from ...consts import DieColor
from ...struct import Cost
from ..pyro.yoimiya import Yoimiya as Y_3_8
from ..pyro.yoimiya import RyuukinSaxifrage as RS_3_8
from ..pyro.yoimiya import FireworkFlareUp, NiwabiFireDance


class RyuukinSaxifrage(RS_3_8):
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3,
        charge = 2
    )


class Yoimiya_3_3(Y_3_8):
    version: Literal['3.3']
    max_charge: int = 2
    skills: List[FireworkFlareUp | NiwabiFireDance | RyuukinSaxifrage] = []

    def _init_skills(self) -> None:
        self.skills = [
            FireworkFlareUp(),
            NiwabiFireDance(),
            RyuukinSaxifrage()
        ]



from typing import Any, List, Literal

from ...event import RoundEndEventArguments

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageIncreaseValue, DamageValue

from ...action import ActionTypes, Actions, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DamageType, DieColor, ElementType, FactionType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, PhysicalNormalAttackBase, 
    CharactorBase, SkillTalent
)


# Summons


class DandelionField(AttackerSummonBase):
    name: Literal['Dandelion Field'] = 'Dandelion Field'
    desc: str = (
        'End Phase: Deal 2 Anemo DMG, heal your active character for 1 HP.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.ANEMO
    damage: int = 2

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        """
        If found Jean with talent, increase anemo damage by 1
        """
        if (
            value.damage_type != DamageType.DAMAGE
            or value.damage_elemental_type != DamageElementalType.ANEMO
            or value.position.player_idx != self.position.player_idx
            or value.target_position.player_idx == self.position.player_idx
        ):
            # not anemo damage, or not from our side attack enemy
            return value
        # find jean with talent
        charactors = match.player_tables[self.position.player_idx].charactors
        for charactor in charactors:
            if charactor.name == 'Jean' and charactor.talent is not None:
                # found, increase damage
                assert mode == 'REAL'
                value.damage += 1
                return value
        return value

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        ret = super().event_handler_ROUND_END(event, match)
        our_active = match.player_tables[
            self.position.player_idx].get_active_charactor()
        ret.append(MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = self.position.player_idx,
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = our_active.position,
                    damage = -1,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = Cost(),
                )
            ],
        ))
        return ret


# Skills


class GaleBlade(ElementalSkillBase):
    name: Literal['Gale Blade'] = 'Gale Blade'
    desc: str = (
        'Deals 3 Anemo DMG, the target is forcibly switched to the next '
        'character.'
    )
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        next_idx = match.player_tables[
            1 - self.position.player_idx].next_charactor_idx()
        if next_idx is None:
            return ret
        # change charactor
        attack_action = ret[0]
        assert attack_action.type == ActionTypes.MAKE_DAMAGE
        attack_action.do_charactor_change = True
        attack_action.charactor_change_idx = next_idx
        return ret


class DandelionBreeze(ElementalBurstBase):
    name: Literal['Dandelion Breeze'] = 'Dandelion Breeze'
    desc: str = (
        'Heals all your characters for 2 HP, summons 1 Dandelion Field.'
    )
    damage: int = -2
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 4,
        charge = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret: List[Actions] = [self.charge_self(-3)]
        charactors = match.player_tables[self.position.player_idx].charactors
        heal_action = MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = self.position.player_idx,
            damage_value_list = [],
        )
        for charactor in charactors:
            if charactor.is_alive:
                heal_action.damage_value_list.append(
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.HEAL,
                        target_position = charactor.position,
                        damage = self.damage,
                        damage_elemental_type = DamageElementalType.HEAL,
                        cost = self.cost.copy(),
                    )
                )
        ret.append(heal_action)
        ret.append(self.create_summon('Dandelion Field'))
        return ret


# Talents


class LandsOfDandelion(SkillTalent):
    name: Literal['Lands of Dandelion']
    desc: str = (
        'Combat Action: When your active character is Jean, equip this card. '
        'After Jean equips this card, immediately use Dandelion Breeze once. '
        'When your Jean, who has this card equipped, is on the field, '
        'Dandelion Field will cause you to deal +1 Anemo DMG.'
    )
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Jean'] = 'Jean'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 4,
        charge = 3
    )
    skill: DandelionBreeze = DandelionBreeze()


# charactor base


class Jean(CharactorBase):
    name: Literal['Jean']
    version: Literal['3.3'] = '3.3'
    desc: str = '''"Dandelion Knight" Jean'''
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | GaleBlade | DandelionBreeze
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.SWORD
    talent: LandsOfDandelion | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Favonius Bladework',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            GaleBlade(),
            DandelionBreeze(),
        ]



from typing import List, Literal
from pydantic import validator
from ..consts import (
    ElementType, FactionType, WeaponType, DamageElementalType,
    ELEMENT_TO_DAMAGE_TYPE
)
from .charactor_base import (
    PhysicalNormalAttackBase, ElementalSkillBase, ElementalBurstBase, 
    CharactorBase
)


class Mob(CharactorBase):
    """
    Mobs. They cannot carry weapons in default. Their normal 
    attacks are 2 physical DMG, elemental skills are 3 element DMG, 
    elemental bursts are 2 charges, 3 cost, and 5 element DMG.
    Their names should be BlablaMob, here Blabla should be one of elemental
    types.
    """
    name: Literal[
        'CryoMob',
        'HydroMob',
        'PyroMob',
        'ElectroMob',
        'GeoMob',
        'DendroMob',
        'AnemoMob',
    ]
    desc: str = 'A _NAME_.'
    version: Literal['1.0'] = '1.0'
    element: ElementType = ElementType.NONE
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | ElementalSkillBase | ElementalBurstBase
    ] = []
    faction: List[FactionType] = []
    weapon_type: WeaponType = WeaponType.OTHER

    @validator('element')
    def element_fits_name(cls, v: ElementType, values, **kwargs):
        """
        Check if element type fits name.
        """
        if 'name' not in values:
            raise AssertionError('Name not found.')
        type_in_name: str = values['name'][:-3].upper()
        element_type: str = v.value
        if type_in_name != element_type:
            raise ValueError(
                f'Element type {element_type} does not fit '
                f'name {type_in_name}.'
            )
        return v

    def __init__(self, **kwargs):
        super().__init__(**kwargs)  # type: ignore
        self.desc = self.desc.replace('_NAME_', self.name)

    def _init_skills(self):
        if self.element == ElementType.NONE:
            # set element by name
            element_name = self.name[:-3].upper()
            self.element = ElementType(element_name)
        element_name = self.element.value.lower()
        element_name = element_name[0].upper() + element_name[1:]
        normal_attack = PhysicalNormalAttackBase(
            name = 'Physical Normal Attack',
            damage_type = DamageElementalType.PHYSICAL,
            cost = PhysicalNormalAttackBase.get_cost(self.element),
        )
        elemental_skill = ElementalSkillBase(
            name = f'{element_name} Elemental Skill',
            damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
            cost = ElementalSkillBase.get_cost(self.element),
        )
        elemental_burst = ElementalBurstBase(
            name = f'{element_name} Elemental Burst',
            damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
            cost = ElementalBurstBase.get_cost(self.element, 3, 2),
            damage = 5,
        )
        self.skills = [normal_attack, elemental_skill, elemental_burst]




from typing import Any, List, Literal

from ...event import SkillEndEventArguments

from ...consts import (
    ELEMENT_TO_ATK_UP_ICON, ELEMENT_TO_DAMAGE_TYPE, CostLabels, DamageType, 
    ElementType, IconType, ObjectPositionType, SkillType
)

from ...action import CreateObjectAction, RemoveObjectAction

from ...modifiable_values import CostValue, DamageIncreaseValue

from .base import RoundTeamStatus, UsageTeamStatus


class Stormzone(UsageTeamStatus):
    name: Literal['Stormzone'] = 'Stormzone'
    desc: str = (
        'When you perform "Switch Character": Spend 1 less Elemental Die. '
        'Usage(s): 2'
    )
    version: Literal['3.7'] = '3.7'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    talent_activated: bool = False
    decrease_cost_success: bool = False
    updated_desc: str = (
        'When you perform "Switch Character": Spend 1 less Elemental Die. '
        'After this effect is triggered, your next Normal Attack this '
        'round will cost 1 less Unaligned Element. Usage(s): 2'
    )

    def renew(self, new_status: 'Stormzone') -> None:
        super().renew(new_status)
        if new_status.talent_activated:
            self.talent_activated = True
            self.desc = self.updated_desc

    def __init__(self, *argv, **kwargs) -> None:
        super().__init__(*argv, **kwargs)
        if self.talent_activated:
            # change desc
            self.desc = self.updated_desc

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        Decrease sw_char cost. 
        """
        if value.position.player_idx != self.position.player_idx:
            # not this player, do nothing
            return value
        if value.cost.label & CostLabels.SWITCH_CHARACTOR.value == 0:
            # not sw_char, do nothing
            return value
        if value.cost.decrease_cost(None):
            # decrease success
            if mode == 'REAL':
                self.decrease_cost_success = True
                self.usage -= 1
        return value

    def event_handler_SWITCH_CHARACTOR(
        self, event: ..., 
        match: Any
    ) -> List[CreateObjectAction | RemoveObjectAction]:
        """
        if this status is talent activated, them create Winds of Harmony.
        Check if need to remove self.
        """
        if not self.decrease_cost_success:
            # not decrease cost success, do nothing
            return []
        self.decrease_cost_success = False
        ret: List[CreateObjectAction | RemoveObjectAction] = []
        if self.talent_activated:
            # create Winds of Harmony
            ret.append(CreateObjectAction(
                object_name = 'Winds of Harmony',
                object_position = self.position,
                object_arguments = {}
            ))
        ret += self.check_should_remove()
        return ret


# Usage status, will not disappear until usage is 0
class WindsOfHarmony(RoundTeamStatus):
    name: Literal['Winds of Harmony'] = 'Winds of Harmony'
    desc: str = (
        "During this round, your charactor's next Normal Attack costs 1 less "
        "Unaligned Element."
    )
    version: Literal['3.7'] = '3.7'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode = Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        Decrease normal attack cost.
        """
        if self.position.check_position_valid(
            value.position, match, player_idx_same = True, 
            target_area = ObjectPositionType.SKILL,
        ):
            # this charactor use skill, check if normal attack
            skill = match.get_object(value.position)
            skill_type: SkillType = skill.skill_type
            if skill_type == SkillType.NORMAL_ATTACK:
                # normal attack, decrease cost
                if value.cost.decrease_cost(None):
                    # decrease success
                    if mode == 'REAL':
                        self.usage -= 1
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        return self.check_should_remove()


class PoeticsOfFuubutsu(UsageTeamStatus):
    name: Literal[
        'Poetics of Fuubutsu: Pyro',
        'Poetics of Fuubutsu: Hydro',
        'Poetics of Fuubutsu: Electro',
        'Poetics of Fuubutsu: Cryo',
    ]
    desc: str = (
        'your Characters and Summons will deal +1 DMG for _ELEMENT_ DMG.'
    )
    version: Literal['3.8'] = '3.8'
    usage: int = 2
    max_usage: int = 2
    element: ElementType = ElementType.NONE
    icon_type: Literal[
        IconType.ATK_UP_FIRE,
        IconType.ATK_UP_WATER,
        IconType.ATK_UP_ELEC,
        IconType.ATK_UP_ICE,
        IconType.ATK_UP
    ] = IconType.ATK_UP

    def __init__(self, *argv, **kwargs) -> None:
        super().__init__(*argv, **kwargs)
        element_name = self.name.split()[-1]
        element = ElementType[element_name.upper()]
        self.element = element
        self.desc = self.desc.replace('_ELEMENT_', element.name.capitalize())
        self.icon_type = ELEMENT_TO_ATK_UP_ICON[element]  # type: ignore

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        """
        If our charactor skill or summon deal corresponding elemental DMG, 
        increase DMG.
        """
        if value.damage_type != DamageType.DAMAGE:
            # not damage, do nothing
            return value
        if value.position.player_idx != self.position.player_idx:
            # not this player, do nothing
            return value
        if value.damage_elemental_type != ELEMENT_TO_DAMAGE_TYPE[self.element]:
            # not corresponding elemental DMG, do nothing
            return value
        if value.position.area not in [
            ObjectPositionType.SKILL, ObjectPositionType.SUMMON
        ]:
            # not charactor or summon, do nothing
            return value
        if value.damage_from_element_reaction:
            # damage from elemental reaction, do nothing
            return value
        # increase DMG
        assert mode == 'REAL'
        self.usage -= 1
        value.damage += 1
        return value


AnemoTeamStatus = Stormzone | WindsOfHarmony | PoeticsOfFuubutsu



from typing import Any, List, Literal

from .base import ArtifactBase

from ....consts import DieColor, ObjectPositionType

from ....struct import Cost, ObjectPosition

from ....event import CharactorDefeatedEventArguments
from ....action import CreateDiceAction, Actions


class GamblersEarrings(ArtifactBase):
    name: Literal["Gambler's Earrings"]
    desc: str = (
        'After an opposing character is defeated: If the character this card '
        'is attached to is the active character, create Omni Element x2. '
        '(Can happen 3 times per match)'
    )
    version: Literal['3.8'] = '3.8'
    cost: Cost = Cost(same_dice_number = 1)
    usage: int = 3

    def equip(self, match: Any) -> List[Actions]:
        """
        Equip this artifact. Reset usage.
        """
        self.usage = 3
        return []

    def event_handler_CHARACTOR_DEFEATED(
        self, event: CharactorDefeatedEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        When an opposing character is defeated, check if the character this 
        card is attached to is the active character. If so, create Omni 
        Element x2.
        """
        target_position = ObjectPosition(
            player_idx = event.action.player_idx,
            # all the following are not used to check, no need to set correctly
            area = ObjectPositionType.INVALID,
            id = -1,
        )
        if not self.position.check_position_valid(
            target_position, match, 
            player_idx_same = False,
            source_area = ObjectPositionType.CHARACTOR,
            source_is_active_charactor = True,
        ):
            # our charactor defeated, or self not active, or self not equipped
            return []
        if self.usage <= 0:
            # no usage left
            return []
        self.usage -= 1
        return [CreateDiceAction(
            player_idx = self.position.player_idx,
            number = 2,
            color = DieColor.OMNI
        )]


Gamblers = GamblersEarrings | GamblersEarrings



from .mona import Mona, ProphecyOfSubmersion, Reflection
from .barbara import Barbara, MelodyLoop, GloriousSeason
from .rhodeia import RhodeiaOfLoch, StreamingSurge, Frog, Squirrel, Raptor
from .xingqiu import Xingqiu, TheScentRemained
from .tartaglia import Tartaglia, AbyssalMayhemHydrospout
from .sangonomiya_kokomi import (
    SangonomiyaKokomi, BakeKurage, TamakushiCasket
)
from .candace import Candace, TheOverflow
from .mirror_maiden import MirrorMaiden, MirrorCage
from .kamisato_ayato import KamisatoAyato, GardenOfPurity, KyoukaFuushi


HydroCharactors = (
    Barbara | Xingqiu | Mona | Tartaglia | SangonomiyaKokomi | KamisatoAyato 
    | Candace
    # finally monsters
    | RhodeiaOfLoch | MirrorMaiden
)
SummonsOfHydroCharactors = (
    MelodyLoop | Reflection | BakeKurage | GardenOfPurity
    # finally monsters
    | Frog | Squirrel | Raptor
)
HydroCharactorTalents = (
    GloriousSeason | TheScentRemained | ProphecyOfSubmersion
    | AbyssalMayhemHydrospout | TamakushiCasket | KyoukaFuushi | TheOverflow
    # finally monsters
    | StreamingSurge | MirrorCage
)



from typing import Any, List, Literal

from ...action import Actions, CreateObjectAction, RemoveObjectAction
from ...struct import Cost

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE, DamageElementalType, DieColor, ElementType, 
    FactionType, ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    CharactorBase, SkillTalent
)


# Skills


class InfluxBlast(ElementalSkillBase):
    name: Literal['Influx Blast'] = 'Influx Blast'
    desc: str = (
        'Deals 2 Hydro DMG. The target character receives Refraction.'
    )
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        First remove existing Refraction on opposite, then create new
        Refraction.
        """
        target_charactors = match.player_tables[
            1 - self.position.player_idx].charactors
        exist_status = None
        for charactor in target_charactors:
            for status in charactor.status:
                if status.name == 'Refraction':
                    exist_status = status
                    break
            if exist_status is not None:
                break
        ret: List[Actions] = []
        if exist_status:
            ret.append(RemoveObjectAction(
                object_position = exist_status.position
            ))
        target_charactor = match.player_tables[
            1 - self.position.player_idx].get_active_charactor()
        args = {}
        if self.is_talent_equipped(match):
            args = {
                'usage': 3,
                'max_usage': 3,
                'is_talent_activated': True,
            }
        return super().get_actions(match) + ret + [
            CreateObjectAction(
                object_position = target_charactor.position.set_area(
                    ObjectPositionType.CHARACTOR_STATUS
                ),
                object_name = 'Refraction',
                object_arguments = args
            )
        ]


# Talents


class MirrorCage(SkillTalent):
    name: Literal['Mirror Cage']
    desc: str = (
        'Combat Action: When your active character is Mirror Maiden, equip '
        'this card. After Mirror Maiden equips this card, immediately use '
        'Influx Blast once. When your Mirror Maiden, who has this card '
        'equipped, creates a Refraction, it will have the following effects: '
        'Starting Duration (Rounds) +1, will increase the Elemental Dice Cost '
        'of switching from a character to which this is attached to another '
        'character by 1.'
    )
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Mirror Maiden'] = 'Mirror Maiden'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 4
    )
    skill: InfluxBlast = InfluxBlast()

# charactor base


class MirrorMaiden(CharactorBase):
    name: Literal['Mirror Maiden']
    version: Literal['3.7'] = '3.7'
    desc: str = '''"Snezhnayan Maiden" Mirror Envoy'''
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | InfluxBlast | ElementalBurstBase
    ] = []
    faction: List[FactionType] = [
        FactionType.FATUI
    ]
    weapon_type: WeaponType = WeaponType.OTHER
    talent: MirrorCage | None = None

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Water Ball',
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            InfluxBlast(),
            ElementalBurstBase(
                name = 'Rippled Reflection',
                damage = 5,
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalBurstBase.get_cost(self.element, 3, 2)
            )
        ]



from .fatui_pyro_agent import FatuiPyroAgent, PaidinFull
from .klee import Klee, PoundingSurprise
from .bennett import Bennett, GrandExpectation
from .yoimiya import Yoimiya, NaganoharaMeteorSwarm
from .xiangling import Xiangling, Guoba, Crossfire
from .yanfei import Yanfei, RightOfFinalInterpretation
from .hu_tao import HuTao, SanguineRouge
from .diluc import Diluc, FlowingFlame
from .amber import Amber, BaronBunny, BunnyTriggered
from .abyss_lector_fathomless_flames import (
    AbyssLectorFathomlessFlames, DarkfireFurnace, EmbersRekindled
)
from .dehya import Dehya, FierySanctumField, StalwartAndTrue


PyroCharactors = (
    Diluc | Xiangling | Bennett | Amber | Yoimiya | Klee | HuTao | Yanfei
    | Dehya
    # finally monsters
    | FatuiPyroAgent | AbyssLectorFathomlessFlames
)
SummonsOfPyroCharactors = (
    Guoba | BaronBunny | FierySanctumField
    # finally monsters
    | DarkfireFurnace
)
PyroCharactorTalents = (
    FlowingFlame | Crossfire | GrandExpectation | BunnyTriggered 
    | NaganoharaMeteorSwarm | PoundingSurprise | SanguineRouge 
    | RightOfFinalInterpretation | StalwartAndTrue
    # finally monsters
    | PaidinFull | EmbersRekindled
)



from typing import Any, List, Literal

from ...summon.base import AttackerSummonBase

from ...modifiable_values import CostValue, DamageValue
from ...event import RoundEndEventArguments, RoundPrepareEventArguments

from ...action import (
    Actions, MakeDamageAction
)
from ...struct import Cost

from ...consts import (
    CostLabels, DamageElementalType, DamageType, DieColor, ElementType, 
    FactionType, ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    CharactorBase, SkillTalent
)


class MelodyLoop(AttackerSummonBase):
    name: Literal['Melody Loop'] = 'Melody Loop'
    desc: str = (
        'End Phase: Heal all your characters for 1 HP and your active '
        'character gains Hydro Application. '
        'Usage(s): 2'
    )
    version: Literal['3.3'] = '3.3'
    damage_elemental_type: DamageElementalType = DamageElementalType.HEAL
    damage: int = -1
    usage: int = 2
    max_usage: int = 2

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        player_idx = self.position.player_idx
        assert self.usage > 0
        self.usage -= 1
        target_table = match.player_tables[player_idx]
        damage_action = MakeDamageAction(
            source_player_idx = player_idx,
            target_player_idx = player_idx,
            damage_value_list = [],
        )
        for cid, charactor in enumerate(target_table.charactors):
            if charactor.is_alive:
                damage_action.damage_value_list.append(
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.HEAL,
                        target_position = charactor.position,
                        damage = self.damage,
                        damage_elemental_type 
                        = self.damage_elemental_type,
                        cost = Cost(),
                    )
                )
            if target_table.active_charactor_idx == cid:
                assert charactor.is_alive
                damage_action.damage_value_list.append(
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.ELEMENT_APPLICATION,
                        target_position = charactor.position,
                        damage = 0,
                        damage_elemental_type
                        = DamageElementalType.HYDRO,
                        cost = Cost(),
                    )
                )
        return [damage_action]


class LetTheShowBegin(ElementalSkillBase):
    name: Literal['Let the Show Begin♪'] = 'Let the Show Begin♪'
    desc: str = (
        'Deals 1 Hydro DMG, summons 1 Melody Loop.'
    )
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_number = 3,
        elemental_dice_color = DieColor.HYDRO
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        ret.append(self.create_summon('Melody Loop'))
        return ret


class ShiningMiracle(ElementalBurstBase):
    name: Literal['Shining Miracle'] = 'Shining Miracle'
    desc: str = '''Heals all of your characters for 4 HP.'''
    damage: int = -4
    damage_type: DamageElementalType = DamageElementalType.HEAL
    cost: Cost = Cost(
        elemental_dice_number = 3,
        elemental_dice_color = DieColor.HYDRO,
        charge = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret: List[Actions] = [self.charge_self(-3)]
        charactors = match.player_tables[self.position.player_idx].charactors
        heal_action = MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = self.position.player_idx,
            damage_value_list = [],
        )
        for charactor in charactors:
            if charactor.is_alive:
                heal_action.damage_value_list.append(
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.HEAL,
                        target_position = charactor.position,
                        damage = self.damage,
                        damage_elemental_type = self.damage_type,
                        cost = self.cost.copy(),
                    )
                )
        ret.append(heal_action)
        return ret


class GloriousSeason(SkillTalent):
    name: Literal['Glorious Season'] = 'Glorious Season'
    desc: str = (
        'Combat Action: When your active character is Barbara, equip this '
        'card. '
        'After Barbara equips this card, immediately use Let the Show Begin♪ '
        'once. When your Barbara, who has this card equipped, is on the '
        'field, Melody Loop will allow you to spend 1 less Elemental Die the '
        'next time you use "Switch Character." (Once per Round)'
    )
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Barbara'] = 'Barbara'
    cost: Cost = Cost(
        elemental_dice_number = 4,
        elemental_dice_color = DieColor.HYDRO,
    )
    skill: LetTheShowBegin = LetTheShowBegin()
    usage: int = 1

    def event_handler_ROUND_PREPARE(
        self, match: Any, event_args: RoundPrepareEventArguments
    ) -> List[Actions]:
        """reset usage"""
        self.usage = 1
        return []

    def value_modifier_COST(
        self, value: CostValue, 
        match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> CostValue:
        """
        Once per round, if summon "Melody Loop" is valid in our summon area,
        and our do switch, and have cost, then cost -1.
        """
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped, no effect
            return value
        if self.usage <= 0:
            # no usage, no effect
            return value
        if value.cost.label & CostLabels.SWITCH_CHARACTOR == 0:
            # not switch charatctor, no effect
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player, no effect
            return value
        summons = match.player_tables[self.position.player_idx].summons
        have_summon = False
        for summon in summons:
            if summon.name == 'Melody Loop':
                have_summon = True
                break
        if not have_summon:
            # not have summon, no effect
            return value
        # decrease 1 any cost
        if value.cost.decrease_cost(None):
            # decrease success
            if mode == 'REAL':
                self.usage -= 1
        return value


class Barbara(CharactorBase):
    name: Literal['Barbara']
    version: Literal['3.3'] = '3.3'
    desc: str = '''"Shining Idol" Barbara'''
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        ElementalNormalAttackBase | LetTheShowBegin | ShiningMiracle
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.CATALYST
    talent: GloriousSeason | None = None

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Whisper of Water',
                damage_type = DamageElementalType.HYDRO,
                cost = ElementalNormalAttackBase.get_cost(
                    ElementType.HYDRO
                )
            ),
            LetTheShowBegin(),
            ShiningMiracle(),
        ]



from typing import Literal

from ..consts import DamageElementalType
from .base import AttackerSummonBase


class CryoHilichurlShooter(AttackerSummonBase):
    name: Literal['Cryo Hilichurl Shooter'] = 'Cryo Hilichurl Shooter'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1


class ElectroHilichurlShooter(AttackerSummonBase):
    name: Literal['Electro Hilichurl Shooter'] = 'Electro Hilichurl Shooter'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1


class HilichurlBerserker(AttackerSummonBase):
    name: Literal['Hilichurl Berserker'] = 'Hilichurl Berserker'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = 1


class HydroSamachurl(AttackerSummonBase):
    name: Literal['Hydro Samachurl'] = 'Hydro Samachurl'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 1


EventSummons = (
    CryoHilichurlShooter | ElectroHilichurlShooter | HilichurlBerserker
    | HydroSamachurl
)



from typing import Any, List, Literal

from ...event import CreateObjectEventArguments, RoundPrepareEventArguments

from ...summon.base import AttackerSummonBase

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class CuileinAnbar(AttackerSummonBase):
    name: Literal['Cuilein-Anbar'] = 'Cuilein-Anbar'
    desc: str = (
        'End Phase: Deal 2 Dendro DMG. '
        'Usage(s): 2'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.DENDRO
    damage: int = 2


# Skills


class FloralBrush(ElementalSkillBase):
    name: Literal['Floral Brush'] = 'Floral Brush'
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3,
    )
    create_status_counter: int = 0

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        In prepare, reset counter
        """
        self.create_status_counter = 0
        return []

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[Actions]:
        """
        If our charactor create Floral Sidewinder, increase create status
        """
        if event.action.object_position.player_idx != self.position.player_idx:
            # not our charactor
            return []
        if event.action.object_name != 'Floral Sidewinder':
            # not Floral Sidewinder
            return []
        self.create_status_counter += 1
        return []

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If talent equipped, before attack, create status.
        """
        ret = super().get_actions(match)
        if self.create_status_counter > 0:
            # created this round, not create
            return ret
        charactor = match.player_tables[
            self.position.player_idx].charactors[self.position.charactor_idx]
        if charactor.talent is None:
            # no talent equipped
            return ret
        ret = [self.create_team_status('Floral Sidewinder')] + ret
        return ret


class TrumpCardKitty(ElementalBurstBase):
    name: Literal['Trump-Card Kitty'] = 'Trump-Card Kitty'
    desc: str = '''Deals 2 Dendro DMG, summons 1 Cuilein-Anbar.'''
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3,
        charge = 2,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_summon('Cuilein-Anbar')
        ]


# Talents


class FloralSidewinder(SkillTalent):
    """
    Because before using skill, the status should appear, and status need to
    listen DAMAGE_RECEIVE event, we should modify both actions of 
    elemental skill and equipping this card.
    """
    name: Literal['Floral Sidewinder']
    desc: str = (
        'Combat Action: When your active character is Collei, equip this '
        'card. After Collei equips this card, immediately use Floral Brush '
        'once. After your Collei, who has this card equipped, uses '
        "Floral Brush, during this Round, when your characters' Skills "
        'trigger Dendro-Related Reactions: Deal 1 Dendro DMG. (Once per Round)'
    )
    version: Literal['3.4'] = '3.4'
    charactor_name: Literal['Collei'] = 'Collei'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 4
    )
    skill: FloralBrush = FloralBrush()


# charactor base


class Collei(CharactorBase):
    name: Literal['Collei']
    version: Literal['3.3'] = '3.3'
    desc: str = '''"Sprout of Rebirth" Collei'''
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | FloralBrush | TrumpCardKitty
    ] = []
    faction: List[FactionType] = [
        FactionType.SUMERU
    ]
    weapon_type: WeaponType = WeaponType.BOW
    talent: FloralSidewinder | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = "Supplicant's Bowmanship",
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            FloralBrush(),
            TrumpCardKitty()
        ]


# finally, modify server/charactor/(element)/__init__.py



from typing import Any, List, Literal

from ....consts import (
    ELEMENT_TO_DIE_COLOR, ElementalReactionType, ObjectPositionType
)

from ....action import CreateDiceAction

from ....event import ReceiveDamageEventArguments, SkillEndEventArguments

from ....struct import Cost
from .base import RoundEffectArtifactBase


class InstructorsCap(RoundEffectArtifactBase):
    name: Literal["Instructor's Cap"]
    desc: str = (
        'After a character triggers an Elemental Reaction: Create 1 Elemental '
        "Die that matches this Character's Elemental Type. (Max 3 times per "
        'Round)'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(any_dice_number = 2)
    max_usage_per_round: int = 3
    element_reaction_triggered: bool = False

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        Record if elemental reaction triggered
        """
        if event.elemental_reaction != ElementalReactionType.NONE:
            # has elemental reaction
            self.element_reaction_triggered = True
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        if elemental reaction reiggered, and self use skill, and have usage,
        create one die with same element type as self.
        """
        if not self.element_reaction_triggered:
            # no elemental reaction triggered
            return []
        self.element_reaction_triggered = False
        if self.usage <= 0:
            # no usage
            return []
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, 
            source_area = ObjectPositionType.CHARACTOR,
            target_area = ObjectPositionType.SKILL
        ):
            # not self player use skill or not equipped
            return []
        # create die
        self.usage -= 1
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        return [
            CreateDiceAction(
                player_idx = self.position.player_idx,
                number = 1,
                color = ELEMENT_TO_DIE_COLOR[charactor.element],
            )
        ]



from typing import List, Literal

from ..hydro.xingqiu import Xingqiu as X_4_1
from ..hydro.xingqiu import Raincutter as R_4_1
from ..hydro.xingqiu import GuhuaStyle, FatalRainscreen


class Raincutter(R_4_1):
    name: Literal['Raincutter'] = 'Raincutter'
    desc: str = (
        'Deals 1 Hydro DMG, grants this character Hydro Application, creates '
        '1 Rainbow Bladework.'
    )
    damage: int = 1


class Xingqiu_3_6(X_4_1):
    version: Literal['3.6']
    skills: List[GuhuaStyle | FatalRainscreen | Raincutter] = []

    def _init_skills(self) -> None:
        self.skills = [
            GuhuaStyle(),
            FatalRainscreen(),
            Raincutter()
        ]



from .noelle import Noelle, IGotYourBack
from .arataki_itto import AratakiItto, Ushi, AratakiIchiban
from .stonehide_lawachurl import StonehideLawachurl, StonehideReforged
from .ningguang import Ningguang, StrategicReserve
from .zhongli import Zhongli, StoneStele, DominanceOfEarth
from .albedo import Albedo, SolarIsotoma, DescentOfDivinity


GeoCharactors = (
    Ningguang | Noelle | Zhongli | Albedo | AratakiItto
    # finally monsters
    | StonehideLawachurl
)
SummonsOfGeoCharactors = StoneStele | SolarIsotoma | Ushi
GeoCharactorTalents = (
    StrategicReserve | IGotYourBack | DominanceOfEarth | DescentOfDivinity 
    | AratakiIchiban 
    # finally monsters
    | StonehideReforged
)



from typing import Any, List, Literal

from ....action import Actions

from ....struct import Cost
from .element_artifacts import SmallElementalArtifact as SEA_4_0
from .element_artifacts import BigElementalArtifact as BEA_4_0
from .gamblers import GamblersEarrings as GE_3_8
from .vermillion_shimenawa import ThunderingPoise as TP_4_0
from .vermillion_shimenawa import VermillionHereafter as VH_4_0
from .vermillion_shimenawa import CapriciousVisage as CV_4_0
from .vermillion_shimenawa import ShimenawasReminiscence as SR_4_0
from .emblem_of_severed_fate import OrnateKabuto as OK_4_0
from .emblem_of_severed_fate import EmblemOfSeveredFate as EOSF_4_1


class SmallElementalArtifact_3_3(SEA_4_0):
    version: Literal['3.3']
    cost: Cost = Cost(same_dice_number = 2)


class BigElementalArtifact_3_6(BEA_4_0):
    version: Literal['3.6']
    cost: Cost = Cost(any_dice_number = 3)


class BigElementalArtifact_3_3(BEA_4_0):
    version: Literal['3.3']
    cost: Cost = Cost(same_dice_number = 3)


class GamblersEarrings_3_3(GE_3_8):
    version: Literal['3.3']
    usage: int = 999

    def equip(self, match: Any) -> List[Actions]:
        """
        Equip this artifact. Reset usage.
        """
        self.usage = 999
        return []


# VermillionShimenawas


class ThunderingPoise(TP_4_0):
    version: Literal['3.7']
    cost: Cost = Cost(same_dice_number = 2)


class VermillionHereafter(VH_4_0):
    version: Literal['3.7']
    cost: Cost = Cost(same_dice_number = 3)


class CapriciousVisage(CV_4_0):
    version: Literal['3.7']
    cost: Cost = Cost(same_dice_number = 2)


class ShimenawasReminiscence(SR_4_0):
    version: Literal['3.7']
    cost: Cost = Cost(same_dice_number = 3)


VermillionShimenawas_3_7 = (
    ThunderingPoise | VermillionHereafter
    | CapriciousVisage | ShimenawasReminiscence
)


# EmblemOfSeveredFate


class OrnateKabuto_3_5(OK_4_0):
    version: Literal['3.5']
    cost: Cost = Cost(any_dice_number = 2)


class EmblemOfSeveredFate_4_0(EOSF_4_1):
    version: Literal['4.0']
    desc: str = (
        'After a character uses an Elemental Burst: The character to which '
        'this is attached gains 1 Energy. The DMG dealt by the '
        "character's Elemental Bursts is increased by 2. "
    )
    max_usage_per_round: int = 999


class EmblemOfSeveredFate_3_7(EmblemOfSeveredFate_4_0):
    version: Literal['3.7']
    cost: Cost = Cost(any_dice_number = 3)


OldVersionArtifacts = (
    EmblemOfSeveredFate_4_0

    | VermillionShimenawas_3_7 | EmblemOfSeveredFate_3_7

    | BigElementalArtifact_3_6

    | OrnateKabuto_3_5

    | GamblersEarrings_3_3 | SmallElementalArtifact_3_3 
    | BigElementalArtifact_3_3
)



from typing import Any, List, Literal

from ...summon.base import SwirlChangeSummonBase

from ...event import RoundEndEventArguments

from ...action import Actions, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class Stormeye(SwirlChangeSummonBase):
    name: Literal['Stormeye'] = 'Stormeye'
    desc: str = (
        'End Phase: Deal 2 Anemo DMG. Your opponent switches to: Character '
        'Closest to Your Current Active Character. '
        'After your character or Summon triggers a Swirl reaction: Convert '
        'the Elemental Type of this card and change its DMG dealt to the '
        'element Swirled. (Can only be converted once before leaving the '
        'field)'
    )
    version: Literal['3.7'] = '3.7'
    usage: int = 2
    max_usage: int = 2
    damage: int = 2

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        Need to change enemy active
        """
        ret = super().event_handler_ROUND_END(event, match)
        our_active = match.player_tables[
            self.position.player_idx].active_charactor_idx
        target_charactors = match.player_tables[
            1 - self.position.player_idx].charactors
        target_idx = our_active
        if target_charactors[target_idx].is_defeated:
            # need to choose another charactor
            for idx, charactor in enumerate(target_charactors):
                if not charactor.is_defeated:
                    target_idx = idx
                    break
            else:
                raise AssertionError('No charactor alive')
        assert len(ret) == 1
        ret[0].do_charactor_change = True
        ret[0].charactor_change_idx = target_idx
        return ret


# Skills


class SkywardSonnet(ElementalSkillBase):
    name: Literal['Skyward Sonnet'] = 'Skyward Sonnet'
    desc: str = '''Deals 2 Anemo DMG, creates 1 Stormzone.'''
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        assert charactor.name == 'Venti'
        talent_activated = False
        if self.is_talent_equipped(match):
            talent_activated = True
        return super().get_actions(match) + [
            self.create_team_status(
                'Stormzone', 
                {'talent_activated': talent_activated}
            )
        ]


class WindsGrandOde(ElementalBurstBase):
    name: Literal["Wind's Grand Ode"] = "Wind's Grand Ode"
    desc: str = '''Deals 2 Anemo DMG, summons 1 Stormeye.'''
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        # create summon first, so it can change element immediately.
        return [self.create_summon('Stormeye')] + super().get_actions(match)


# Talents


class EmbraceOfWinds(SkillTalent):
    name: Literal['Embrace of Winds']
    desc: str = (
        'Combat Action: When your active character is Venti, equip this card. '
        'After Venti equips this card, immediately use Skyward Sonnet once. '
        'After a Stormzone created by your Venti, who has this card equipped, '
        'is triggered, the next Normal Attack performed by your character in '
        'this Round will cost 1 less Unaligned Element.'
    )
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Venti'] = 'Venti'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3
    )
    skill: SkywardSonnet = SkywardSonnet()


# charactor base


class Venti(CharactorBase):
    name: Literal['Venti']
    version: Literal['3.7'] = '3.7'
    desc: str = '''"Windborne Bard" Venti'''
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | SkywardSonnet | WindsGrandOde
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.BOW
    talent: EmbraceOfWinds | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Divine Marksmanship',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SkywardSonnet(),
            WindsGrandOde()
        ]


# finally, modify server/charactor/(element)/__init__.py



import logging

from pydantic import validator

from .object_base import CardBases

from .charactor.charactor_base import CharactorBase

from ..utils import BaseModel, get_instance_from_type_unions
from typing import Literal, List
from . import Cards
from . import Charactors


class Deck(BaseModel):
    name: Literal['Deck'] = 'Deck'
    charactors: List[CharactorBase] = []
    cards: List[CardBases] = []

    @validator('charactors', each_item = True, pre = True)
    def parse_charactors(cls, v):
        return get_instance_from_type_unions(Charactors, v)

    @validator('cards', each_item = True, pre = True)
    def parse_cards(cls, v):
        return get_instance_from_type_unions(Cards, v)

    def check_legal(self, card_number: int | None, 
                    max_same_card_number: int | None, 
                    charactor_number: int | None,
                    check_restriction: bool) -> bool:
        """
        check whether the deck is legal.
        1. the number of charactors
        2. the number of cards
        3. the number of cards with the same name
        4. cards with special carrying rules
        """
        if charactor_number is not None:
            if len(self.charactors) != charactor_number:
                logging.error(f'charactor number should be {charactor_number}')
                return False
        if card_number is not None:
            if len(self.cards) != card_number:
                logging.error(f'card number should be {card_number}')
                return False
        if max_same_card_number is not None:
            card_names = [c.name for c in self.cards]
            for card in card_names:
                if card_names.count(card) > max_same_card_number:
                    logging.error(
                        f'card {card} should not be more than '
                        f'{max_same_card_number} in deck'
                    )
                    return False
        if check_restriction:
            for card in self.cards:
                restriction = card.get_deck_restriction()
                if restriction.type == 'NONE':
                    continue
                elif restriction.type == 'FACTION':
                    counter = 0
                    for charactor in self.charactors:
                        if restriction.name in charactor.faction:
                            counter += 1
                    if counter < restriction.number:
                        logging.error(
                            f'to use card {card.name}, '
                            f'charactor number of faction {restriction.name} '
                            f'should be at least {restriction.number}'
                        )
                        return False
                elif restriction.type == 'CHARACTOR':
                    counter = 0
                    for charactor in self.charactors:
                        if restriction.name == charactor.name:
                            counter += 1
                    if counter < restriction.number:
                        logging.error(
                            f'to use card {card.name}, '
                            f'charactor number of name {restriction.name} '
                            f'should be at least {restriction.number}'
                        )
                        return False
                elif restriction.type == 'ELEMENT':
                    counter = 0
                    for charactor in self.charactors:
                        if restriction.name == charactor.element.value:
                            counter += 1
                    if counter < restriction.number:
                        logging.error(
                            f'to use card {card.name}, '
                            f'charactor number of element {restriction.name} '
                            f'should be at least {restriction.number}'
                        )
                        return False
                else:
                    raise NotImplementedError(
                        f'restriction type {restriction.type} not implemented'
                    )
        return True

    @staticmethod
    def from_str(deck_str: str) -> 'Deck':
        """
        Convert deck string to deck object. One line contains one command, 
        card or charactor. 

        To specify default versions for all cards that don't have version
        specified, use 'default_version:version'. If not set, cards without
        version information will not have version in its argument. This line
        must appear before any other lines.

        If declare a charactor, use 'charactor:'. Otherwise 
        write card name directly. 

        To specify different version of cards, add `@version` after the
        line.

        To declare multiple same cards or charactors, add `*number` after 
        the line. When both marks are used, specify version first.
        You can write comment line start with '#'.

        Examples:

        default_version:4.0
        charactor:Fischl
        charactor:Mona
        charactor:Nahida
        # you can write comment line with '#'
        Timmie*15
        Rana*15

        """
        deck_str = deck_str.strip()
        deck = Deck()
        default_version: str | None = None
        for line in deck_str.split('\n'):
            line = line.strip()
            if line == '':
                continue
            if line.startswith('default_version:'):
                default_version = line[16:]
                continue
            if line[0] == '#':
                # comment line
                continue
            # deal with numbers
            if '*' in line:
                line, number = line.split('*')
                number = int(number)
            else:
                number = 1
            # deal with version
            if '@' in line:
                line, version = line.split('@')
                version = version.strip()
            else:
                version = default_version
            if line.startswith('charactor:'):
                args = { 'name': line[10:] }
                if version is not None:
                    args['version'] = version
                for _ in range(number):
                    deck.charactors.append(
                        get_instance_from_type_unions(
                            Charactors, args
                        )
                    )
            else:
                for _ in range(number):
                    args = { 'name': line }
                    if version is not None:
                        args['version'] = version
                    deck.cards.append(
                        get_instance_from_type_unions(
                            Cards, args
                        )
                    )
        return deck

    def to_str(self) -> str:
        """
        Convert deck object to deck string.
        """
        deck_str = ''
        for charactor in self.charactors:
            deck_str += f'charactor:{charactor.name}'
            deck_str += f'@{charactor.version}'
            deck_str += '\n'
        for card in self.cards:
            deck_str += card.name
            deck_str += f'@{card.version}'
            deck_str += '\n'
        return deck_str



from typing import Literal, List, Any

from ....event import MoveObjectEventArguments, RoundPrepareEventArguments

from ....object_base import CardBase
from ....struct import Cost
from ....consts import ObjectType, CostLabels, ObjectPositionType
from ....action import Actions, MoveObjectAction, RemoveObjectAction
from ....struct import ObjectPosition


class ArtifactBase(CardBase):
    """
    Base class of artifacts.
    """
    name: str
    desc: str
    version: str
    cost: Cost
    usage: int

    type: Literal[ObjectType.ARTIFACT] = ObjectType.ARTIFACT
    cost_label: int = CostLabels.CARD.value | CostLabels.ARTIFACT.value
    remove_when_used: bool = False

    def equip(self, match: Any) -> List[Actions]:
        """
        The artifact is equipped, i.e. from hand to charactor. Set the status
        of artifact, and if it has actions when equipped, return the actions.
        """
        # should be used by small qianyan
        return []

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # can quip on all self alive charactors
        ret: List[ObjectPosition] = []
        for charactor in match.player_tables[
                self.position.player_idx].charactors:
            if charactor.is_alive:
                ret.append(charactor.position)
        return ret

    def is_valid(self, match: Any) -> bool:
        # can only be used when in hand
        return self.position.area == ObjectPositionType.HAND

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[MoveObjectAction | RemoveObjectAction]:
        """
        Act the artifact. will place it into artifact area.
        When other artifact is equipped, remove the old one.
        """
        assert target is not None
        ret: List[MoveObjectAction | RemoveObjectAction] = []
        position = target.set_id(self.id)
        charactor_id = target.id
        assert position.area == ObjectPositionType.CHARACTOR
        assert position.player_idx == self.position.player_idx
        charactors = match.player_tables[position.player_idx].charactors
        for charactor in charactors:
            if charactor.id == charactor_id:
                # check if need to remove current artifact
                if charactor.artifact is not None:
                    ret.append(RemoveObjectAction(
                        object_position = charactor.artifact.position,
                    ))
        ret.append(MoveObjectAction(
            object_position = self.position,
            target_position = position,
        ))
        return ret

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        """
        When this artifact is moved from hand to charactor, it is considered
        as equipped, and will call `self.equip`.
        """
        if (
            event.action.object_position.id == self.id
            and event.action.object_position.area == ObjectPositionType.HAND
            and event.action.target_position.area 
            == ObjectPositionType.CHARACTOR
        ):
            # this artifact equipped from hand to charactor
            return self.equip(match)
        return []


class RoundEffectArtifactBase(ArtifactBase):
    """
    Artifacts that has round effects. Refresh their usage when equipped and
    at round preparing stage.
    Instead of setting usage, set max_usage_per_round.
    """
    name: str
    desc: str
    version: str
    cost: Cost
    max_usage_per_round: int 

    usage: int = 0

    def equip(self, match: Any) -> List[Actions]:
        self.usage = self.max_usage_per_round
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        self.usage = self.max_usage_per_round
        return []

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        """
        When this weapon is moved from charactor to charactor, and mark as
        reset_usage, reset usage.
        """
        if (
            event.action.object_position.id == self.id
            and event.action.object_position.area 
            == ObjectPositionType.CHARACTOR
            and event.action.target_position.area 
            == ObjectPositionType.CHARACTOR
            and event.action.reset_usage
        ):
            # this weapon equipped from charactor to charactor
            self.usage = self.max_usage_per_round
        return super().event_handler_MOVE_OBJECT(event, match)



from .system import SystemTeamStatus
from .hydro_charactors import HydroCharactorTeamStatus
from .dendro_charactors import DendroTeamStatus
from .geo_charactors import GeoTeamStatus
from .anemo_charactors import AnemoTeamStatus
from .pyro_charactors import PyroTeamStatus
from .cryo_charactors import CryoTeamStatus
from .electro_charactors import ElectroTeamStatus
from .event_cards import EventCardTeamStatus
from .weapons import WeaponTeamStatus

from .old_version import OldVersionTeamStatus


TeamStatus = (
    SystemTeamStatus | HydroCharactorTeamStatus | DendroTeamStatus
    | GeoTeamStatus | AnemoTeamStatus | PyroTeamStatus | CryoTeamStatus 
    | ElectroTeamStatus | EventCardTeamStatus | WeaponTeamStatus
    # finally, old version status
    | OldVersionTeamStatus
)



from typing import Any, List, Literal


from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageValue
from ...event import SkillEndEventArguments

from ...action import Actions, CharactorReviveAction, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DamageType, DieColor, ElementType, FactionType, 
    ObjectPositionType, SkillType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class HeraldOfFrost(AttackerSummonBase):
    name: Literal['Herald of Frost'] = 'Herald of Frost'
    desc: str = (
        'End Phase: Deal 1 Cryo DMG. '
        'When this Summon is on the field, after Qiqi uses Normal Attack: '
        'Heals your character that took the most DMG for 1 HP.'
    )
    version: Literal['4.0'] = '4.0'
    usage: int = 3
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        if event.action.position.player_idx != self.position.player_idx:
            # not our player skill made damage, do nothing
            return []
        assert event.action.position.area == ObjectPositionType.SKILL
        charactors = match.player_tables[self.position.player_idx].charactors
        charactor = charactors[event.action.position.charactor_idx]
        if charactor.name != 'Qiqi':
            # not Qiqi, do nothing
            return []
        if event.action.skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack, do nothing
            return []
        # select charactor with most damage taken
        selected_charactor = charactor
        for c in charactors:
            if c.is_alive and c.damage_taken > selected_charactor.damage_taken:
                selected_charactor = c
        # heal charactor
        return [MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = self.position.player_idx,
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = selected_charactor.position,
                    damage = -1,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = Cost(),
                )
            ],
        )]


# Skills


class AdeptusArtHeraldOfFrost(ElementalSkillBase):
    name: Literal[
        'Adeptus Art: Herald of Frost'] = 'Adeptus Art: Herald of Frost'
    desc: str = '''Summons 1 Herald of Frost.'''
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object
        """
        return [
            self.charge_self(1),
            self.create_summon('Herald of Frost'),
        ]


class AdeptusArtPreserverOfFortune(ElementalBurstBase):
    name: Literal[
        'Adeptus Art: Preserver of Fortune'
    ] = 'Adeptus Art: Preserver of Fortune'
    desc: str = '''Deals 3 Cryo DMG, creates 1 Fortune-Preserving Talisman.'''
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3,
        charge = 3
    )

    revive_usage: int = 2

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Make damage and create status. Then if has talent and has revive usage,
        revive all defeated charactors with 2 hp.
        """
        ret = super().get_actions(match) + [
            self.create_team_status('Fortune-Preserving Talisman')
        ]
        if not self.is_talent_equipped(match):
            # no talent
            return ret
        # has talent
        if self.revive_usage <= 0:
            # no revive usage
            return ret
        defeated_charactors: List[CharactorBase] = []
        charactors = match.player_tables[self.position.player_idx].charactors
        for c in charactors:
            if c.is_defeated:
                defeated_charactors.append(c)
        if len(defeated_charactors) == 0:
            # no defeated charactor
            return ret
        # has defeated charactor, revive them
        self.revive_usage -= 1
        for c in defeated_charactors:
            ret.append(CharactorReviveAction(
                player_idx = self.position.player_idx,
                charactor_idx = c.position.charactor_idx,
                revive_hp = 2
            ))
        return ret


# Talents


class RiteOfResurrection(SkillTalent):
    name: Literal['Rite of Resurrection']
    desc: str = (
        'Combat Action: When your active character is Qiqi, equip this card. '
        'After Qiqi equips this card, immediately use Adeptus Art: Preserver '
        'of Fortune once. After your Qiqi, who has this card equipped, uses '
        'Adeptus Art: Preserver of Fortune, she revives all your fallen '
        'characters and heals them for 2 HP. (Can happen 2 times per match)'
    )
    version: Literal['4.0'] = '4.0'
    charactor_name: Literal['Qiqi'] = 'Qiqi'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 5,
        charge = 3
    )
    skill: AdeptusArtPreserverOfFortune = AdeptusArtPreserverOfFortune()


# charactor base


class Qiqi(CharactorBase):
    name: Literal['Qiqi']
    version: Literal['4.0'] = '4.0'
    desc: str = '''"Icy Resurrection" Qiqi'''
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | AdeptusArtHeraldOfFrost 
        | AdeptusArtPreserverOfFortune
    ] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.SWORD
    talent: RiteOfResurrection | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Ancient Sword Art',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            AdeptusArtHeraldOfFrost(),
            AdeptusArtPreserverOfFortune(),
        ]



from typing import Literal
from ..consts import IconType, ObjectType
from ..object_base import ObjectBase


class StatusBase(ObjectBase):
    """
    Base class of status.
    """
    name: str
    desc: str
    version: str
    show_usage: bool = True
    usage: int
    max_usage: int
    type: Literal[ObjectType.CHARACTOR_STATUS, ObjectType.TEAM_STATUS]
    renew_type: Literal['ADD', 'RESET', 'RESET_WITH_MAX'] = 'ADD'

    icon_type: IconType

    def renew(self, new_status: 'StatusBase') -> None:
        """
        Renew the status. 
        """
        assert self.renew_type == 'ADD', 'Other types not tested'
        if self.max_usage < new_status.max_usage:
            self.max_usage = new_status.max_usage
        if self.max_usage <= self.usage:
            # currently over maximum, ignore renew.
            return
        self.usage += new_status.usage
        if self.max_usage < self.usage:
            self.usage = self.max_usage
        return
        # elif self.renew_type == 'RESET':
        #     self.usage = new_status.usage
        # elif self.renew_type == 'RESET_WITH_MAX':
        #     raise NotImplementedError('Not tested part')
        #     self.usage = max(self.usage, new_status.usage)



from typing import Any, List, Literal

from ...modifiable_values import DamageValue

from ...action import (
    ChangeObjectUsageAction, ChargeAction, CreateDiceAction, 
    CreateObjectAction, DrawCardAction, GenerateSwitchCardRequestAction, 
    MakeDamageAction, RemoveDiceAction, SwitchCharactorAction
)
from ...consts import (
    ELEMENT_TO_DIE_COLOR, DamageElementalType, DamageType, DieColor, 
    ElementType, FactionType, ObjectPositionType
)

from ...object_base import CardBase
from ...struct import Cost, DeckRestriction, ObjectPosition


class ElementalResonanceCardBase(CardBase):
    element: ElementType
    restriction_desc: str = (
        '(You must have at least 2 XXX characters in your deck to add '
        'this card to your deck.)'
    )

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        restriction = self.restriction_desc.replace(
            'XXX', self.element.name.capitalize()
        )
        self.desc = self.desc.replace(restriction, '')
        self.desc += restriction

    def get_deck_restriction(self) -> DeckRestriction:
        """
        For element resonance cards, should contain at least 2 charactors of
        the corresponding element.
        """
        return DeckRestriction(
            type = 'ELEMENT',
            name = self.element.value,
            number = 2
        )


name_to_element_type = {
    'Elemental Resonance: Woven Flames': ElementType.PYRO,
    'Elemental Resonance: Woven Ice': ElementType.CRYO,
    'Elemental Resonance: Woven Stone': ElementType.GEO,
    'Elemental Resonance: Woven Thunder': ElementType.ELECTRO,
    'Elemental Resonance: Woven Waters': ElementType.HYDRO,
    'Elemental Resonance: Woven Weeds': ElementType.DENDRO,
    'Elemental Resonance: Woven Winds': ElementType.ANEMO,
}


class WovenCards(ElementalResonanceCardBase):
    name: Literal[
        'Elemental Resonance: Woven Flames', 
        'Elemental Resonance: Woven Ice',
        'Elemental Resonance: Woven Stone', 
        'Elemental Resonance: Woven Thunder',
        'Elemental Resonance: Woven Waters',
        'Elemental Resonance: Woven Weeds',
        'Elemental Resonance: Woven Winds'
    ]
    desc: str = '''Create 1 XXX Die.'''
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost()
    element: ElementType = ElementType.NONE  # will update element in __init__

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        self.element = name_to_element_type[self.name]
        self.desc = self.desc.replace('XXX', self.element.name.capitalize())

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateDiceAction]:
        """
        Create one corresponding element dice
        """
        return [CreateDiceAction(
            player_idx = self.position.player_idx,
            color = ELEMENT_TO_DIE_COLOR[self.element],
            number = 1,
        )]


class ShatteringIce(ElementalResonanceCardBase):
    name: Literal['Elemental Resonance: Shattering Ice']
    desc: str = (
        'During this Round, your current active character will deal +2 DMG '
        'for the next instance.'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 1
    )
    element: ElementType = ElementType.CRYO

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # avtive character
        return [match.player_tables[
            self.position.player_idx].get_active_charactor().position]

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Create status
        """
        assert target is not None
        return [CreateObjectAction(
            object_name = self.name,
            object_position = target.set_area(
                ObjectPositionType.CHARACTOR_STATUS),
            object_arguments = {}
        )]


class SoothingWater(ElementalResonanceCardBase):
    name: Literal['Elemental Resonance: Soothing Water']
    desc: str = (
        'Heal your active character for 2 HP and all your characters on '
        'standby for 1 HP.'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 1
    )
    element: ElementType = ElementType.HYDRO

    def is_valid(self, match: Any) -> bool:
        """
        At least one charactor is taken damage.
        """
        for charactor in match.player_tables[
                self.position.player_idx].charactors:
            if charactor.is_alive and charactor.damage_taken > 0:
                return True
        return False

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[MakeDamageAction]:
        """
        heal active 2, others 1.
        """
        active_charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        damage_action = MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = self.position.player_idx,
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = active_charactor.position,
                    damage = -2,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = self.cost.copy()
                )
            ],
        )
        for charactor in match.player_tables[
                self.position.player_idx].charactors:
            if (
                charactor.is_alive
                and charactor.position != active_charactor.position
            ):
                damage_action.damage_value_list.append(
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.HEAL,
                        target_position = charactor.position,
                        damage = -1,
                        damage_elemental_type = DamageElementalType.HEAL,
                        cost = self.cost.copy()
                    )
                )
        return [damage_action]


class FerventFlames(ElementalResonanceCardBase):
    name: Literal['Elemental Resonance: Fervent Flames']
    desc: str = (
        'During this round, the next instance of Pyro-Related Reactions your '
        'current active character triggers deals +3 DMG.'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 1
    )
    element: ElementType = ElementType.PYRO

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # avtive character
        return [match.player_tables[
            self.position.player_idx].get_active_charactor().position]

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Create status
        """
        assert target is not None
        return [CreateObjectAction(
            object_name = self.name,
            object_position = target.set_area(
                ObjectPositionType.CHARACTOR_STATUS),
            object_arguments = {}
        )]


class HighVoltage(ElementalResonanceCardBase):
    name: Literal['Elemental Resonance: High Voltage']
    desc: str = (
        'During this round, one of your characters without maximum Energy '
        'gains 1 Energy. (Active Character prioritized)'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 1
    )
    element: ElementType = ElementType.ELECTRO

    def is_valid(self, match: Any) -> bool:
        """
        At least one charactor without maximum Energy.
        """
        for charactor in match.player_tables[
                self.position.player_idx].charactors:
            if charactor.is_alive and charactor.charge < charactor.max_charge:
                return True
        return False

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChargeAction]:
        """
        active charactor first to gain energy.
        """
        assert target is None
        active_charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        if active_charactor.charge < active_charactor.max_charge:
            return [ChargeAction(
                player_idx = self.position.player_idx,
                charactor_idx = active_charactor.position.charactor_idx,
                charge = 1,
            )]
        for charactor in match.player_tables[
                self.position.player_idx].charactors:
            if (
                charactor.is_alive
                and charactor.charge < charactor.max_charge
            ):
                return [ChargeAction(
                    player_idx = self.position.player_idx,
                    charactor_idx = charactor.position.charactor_idx,
                    charge = 1,
                )]
        else:
            raise AssertionError('No charactor can be charged.')


class ImpetuousWinds(ElementalResonanceCardBase):
    name: Literal['Elemental Resonance: Impetuous Winds']
    desc: str = (
        'Switch to the target character and create Omni Element x1.'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 1
    )
    element: ElementType = ElementType.ANEMO

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # all alive charactors except active charactor
        active_idx = match.player_tables[
            self.position.player_idx].active_charactor_idx
        return [
            charactor.position for (cid, charactor) in enumerate(
                match.player_tables[self.position.player_idx].charactors
            )
            if charactor.is_alive and cid != active_idx
        ]

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateDiceAction | SwitchCharactorAction]:
        """
        Create one omni element dice
        """
        assert target is not None
        return [
            SwitchCharactorAction(
                player_idx = target.player_idx,
                charactor_idx = target.charactor_idx,
            ),
            CreateDiceAction(
                player_idx = self.position.player_idx,
                color = DieColor.OMNI,
                number = 1,
            ),
        ]


class EnduringRock(ElementalResonanceCardBase):
    name: Literal['Elemental Resonance: Enduring Rock']
    desc: str = (
        'During this round, after your character deals Geo DMG next time: '
        'Should there be any Combat Status on your side that provides Shield, '
        'grant one such Status with 3 Shield points.'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 1
    )
    element: ElementType = ElementType.GEO

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Create status
        """
        return [CreateObjectAction(
            object_name = self.name,
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.TEAM_STATUS,
                id = -1,
            ),
            object_arguments = {}
        )]


class SprawlingGreenery(ElementalResonanceCardBase):
    name: Literal['Elemental Resonance: Sprawling Greenery']
    desc: str = (
        'During this round, the next Elemental Reaction you trigger deals '
        '+2 DMG. Your Burning Flame, Dendro Core, and Catalyzing Field gain '
        '+1 Usage(s).'
    )
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 1
    )
    element: ElementType = ElementType.DENDRO

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction | ChangeObjectUsageAction]:
        """
        Create status and add usage if exists. 
        """
        assert target is None
        ret: List[CreateObjectAction | ChangeObjectUsageAction] = [
            CreateObjectAction(
                object_name = self.name,
                object_position = ObjectPosition(
                    player_idx = self.position.player_idx,
                    area = ObjectPositionType.TEAM_STATUS,
                    id = -1,
                ),
                object_arguments = {}
            )
        ]
        object_names = ['Burning Flame', 'Dendro Core', 'Catalyzing Field']
        status = match.player_tables[self.position.player_idx].team_status
        for s in status:
            if s.name in object_names:
                ret.append(ChangeObjectUsageAction(
                    object_position = s.position,
                    change_type = 'DELTA',
                    change_usage = 1,
                ))
        summons = match.player_tables[self.position.player_idx].summons
        for s in summons:
            if s.name in object_names:
                ret.append(ChangeObjectUsageAction(
                    object_position = s.position,
                    change_type = 'DELTA',
                    change_usage = 1,
                ))
        return ret


class NationResonanceCardBase(CardBase):
    faction: FactionType
    restriction_desc: str = (
        '(You must have at least 2 XXX characters in your deck to add '
        'this card to your deck.)'
    )

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        restriction = self.restriction_desc.replace(
            'XXX', self.faction.name.capitalize()
        )
        self.desc = self.desc.replace(restriction, '')
        self.desc += restriction

    def get_deck_restriction(self) -> DeckRestriction:
        """
        For nation resonance cards, should contain at least 2 charactors of
        the corresponding faction.
        """
        return DeckRestriction(
            type = 'FACTION',
            name = self.faction.value,
            number = 2
        )


class WindAndFreedom(NationResonanceCardBase):
    name: Literal['Wind and Freedom']
    desc: str = (
        'In this Round, when an opposing character is defeated during your '
        'Action, you can continue to act again when that Action ends. '
    )
    version: Literal['4.1'] = '4.1'
    cost: Cost = Cost(same_dice_number = 1)
    faction: FactionType = FactionType.MONDSTADT

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Create Wind and Freedom team status.
        """
        assert target is None
        return [CreateObjectAction(
            object_name = self.name,
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.TEAM_STATUS,
                id = -1,
            ),
            object_arguments = {
                'version': self.version
            },
        )]


class StoneAndContracts(NationResonanceCardBase):
    name: Literal['Stone and Contracts']
    desc: str = (
        'When the Action Phase of the next Round begins: Create 3 Omni '
        'Element. '
    )
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(any_dice_number = 3)
    faction: FactionType = FactionType.LIYUE

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        create Stone and Contracts team status.
        """
        assert target is None
        return [CreateObjectAction(
            object_name = self.name,
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.TEAM_STATUS,
                id = -1,
            ),
            object_arguments = {},
        )]


class ThunderAndEternity(NationResonanceCardBase):
    name: Literal['Thunder and Eternity']
    desc: str = (
        'Convert all your Elemental Dice to Omni Element. '
    )
    version: Literal['4.0'] = '4.0'
    cost: Cost = Cost()
    faction: FactionType = FactionType.INAZUMA

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_dice_color(self, match: Any) -> DieColor:
        """
        Return omni color
        """
        return DieColor.OMNI

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[RemoveDiceAction | CreateDiceAction]:
        """
        remove all dice, and create same number dice
        """
        assert target is None
        player_table = match.player_tables[self.position.player_idx]
        dice_number = len(player_table.dice.colors)
        return [
            RemoveDiceAction(
                player_idx = self.position.player_idx,
                dice_idxs = list(range(dice_number)),
            ),
            CreateDiceAction(
                player_idx = self.position.player_idx,
                color = self.get_dice_color(match),
                number = len(player_table.dice.colors),
            )
        ]


class NatureAndWisdom(NationResonanceCardBase):
    name: Literal['Nature and Wisdom']
    desc: str = (
        'Draw 1 card. After that, switch any cards in your hand. '
    )
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 1)
    faction: FactionType = FactionType.SUMERU

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[DrawCardAction | GenerateSwitchCardRequestAction]:
        """
        draw 1 card, and switch any cards in your hand.
        """
        assert target is None
        return [
            DrawCardAction(
                player_idx = self.position.player_idx,
                number = 1,
                draw_if_filtered_not_enough = True
            ),
            GenerateSwitchCardRequestAction(
                player_idx = self.position.player_idx,
            )
        ]


class AbyssalSummons(NationResonanceCardBase):
    name: Literal['Abyssal Summons']
    desc: str = '''Summon 1 Random Hilichurl Summon! '''
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 2)
    faction: FactionType = FactionType.MONSTER

    def is_valid(self, match: Any) -> bool:
        # if no place for summon, return False
        table = match.player_tables[self.position.player_idx]
        return len(table.summons) < match.config.max_summon_number

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Create 1 Random Hilichurl Summon!
        """
        assert target is None
        assert not match.config.recreate_mode, (
            'Currently not support recreate mode.'
        )
        hilichurl_names = [
            'Cryo Hilichurl Shooter',
            'Electro Hilichurl Shooter',
            'Hilichurl Berserker',
            'Hydro Samachurl'
        ]
        # if already has above summons, do not re-create it
        summons = match.player_tables[self.position.player_idx].summons
        for summon in summons:
            if summon.name in hilichurl_names:
                hilichurl_names.remove(summon.name)
        assert len(hilichurl_names) > 0, 'No Hilichurl summon can be created.'
        selected_name = hilichurl_names[int(match._random() 
                                            * len(hilichurl_names))]
        return [CreateObjectAction(
            object_name = selected_name,
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.SUMMON,
                id = -1,
            ),
            object_arguments = {},
        )]


class FatuiConspiracy(NationResonanceCardBase):
    name: Literal['Fatui Conspiracy']
    desc: str = (
        "Create 1 Fatui Ambusher of a random type on the opponent's field."
    )
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 2)
    faction: FactionType = FactionType.FATUI

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Create 1 Random Status
        """
        assert target is None
        assert not match.config.recreate_mode, (
            'Currently not support recreate mode.'
        )
        fatui_names = [
            'Fatui Ambusher: Cryo Cicin Mage',
            'Fatui Ambusher: Mirror Maiden',
            'Fatui Ambusher: Pyroslinger Bracer',
            'Fatui Ambusher: Electrohammer Vanguard'
        ]
        # if enemy already has above status, do not re-create it
        enemy_status = match.player_tables[
            1 - self.position.player_idx].team_status
        for status in enemy_status:
            if status.name in fatui_names:
                fatui_names.remove(status.name)
        assert len(fatui_names) > 0, 'No Fatui status can be created.'
        # create with random name
        selected_name = fatui_names[int(match._random() * len(fatui_names))]
        return [CreateObjectAction(
            object_name = selected_name,
            object_position = ObjectPosition(
                player_idx = 1 - self.position.player_idx,
                area = ObjectPositionType.TEAM_STATUS,
                id = -1,
            ),
            object_arguments = {},
        )]


ElementResonanceCards = (
    WovenCards | ShatteringIce | SoothingWater | FerventFlames | HighVoltage
    | ImpetuousWinds | EnduringRock | SprawlingGreenery
)
NationResonanceCards = (
    WindAndFreedom | StoneAndContracts | ThunderAndEternity | NatureAndWisdom
    | AbyssalSummons | FatuiConspiracy
)



from typing import Any, List, Literal

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    CharactorBase, PhysicalNormalAttackBase, SkillTalent
)


class Breastplate(ElementalSkillBase):
    name: Literal['Breastplate'] = 'Breastplate'
    desc: str = '''Deals 1 Geo DMG, creates 1 Full Plate.'''
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_number = 3,
        elemental_dice_color = DieColor.GEO
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        ret.append(self.create_team_status('Full Plate'))
        return ret


class SweepingTime(ElementalBurstBase):
    name: Literal['Sweeping Time'] = 'Sweeping Time'
    desc: str = '''Deals 4 Geo DMG. This character gains Sweeping Time.'''
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_number = 4,
        elemental_dice_color = DieColor.GEO,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        ret.append(self.create_charactor_status('Sweeping Time'))
        return ret


class IGotYourBack(SkillTalent):
    name: Literal['I Got Your Back']
    desc: str = (
        'Combat Action: When your active character is Noelle, equip this '
        'card. After Noelle equips this card, immediately use Breastplate '
        'once. When your Noelle, who has this card equipped, creates a Full '
        'Plate, it will heal all your characters for 1 HP after Noelle uses '
        'a Normal Attack. (Once per Round)'
    )
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Noelle'] = 'Noelle'
    cost: Cost = Cost(
        elemental_dice_number = 3,
        elemental_dice_color = DieColor.GEO
    )
    skill: Breastplate = Breastplate()


class Noelle(CharactorBase):
    name: Literal['Noelle']
    version: Literal['3.3'] = '3.3'
    desc: str = '''"Chivalric Blossom" Noelle'''
    element: ElementType = ElementType.GEO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | Breastplate | SweepingTime
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.CLAYMORE
    talent: IGotYourBack | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Favonius Bladework - Maid',
                cost = PhysicalNormalAttackBase.get_cost(
                    ElementType.GEO,
                )
            ),
            Breastplate(),
            SweepingTime()
        ]



from .server.interaction import (  # noqa: F401
    Responses, 
    SwitchCardResponse, ChooseCharactorResponse, RerollDiceResponse,
    SwitchCharactorResponse, ElementalTuningResponse,
    DeclareRoundEndResponse, UseSkillResponse, UseCardResponse
)



from typing import Any, List, Literal

from ...modifiable_values import DamageIncreaseValue

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, SkillType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    CharactorBase, SkillTalent
)
# Skills


class SealOfApproval(ElementalNormalAttackBase):
    name: Literal['Seal of Approval'] = 'Seal of Approval'
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = ElementalNormalAttackBase.get_cost(ElementType.PYRO)


class SignedEdict(ElementalSkillBase):
    name: Literal['Signed Edict'] = 'Signed Edict'
    desc: str = (
        'Deals 3 Pyro DMG and attaches Scarlet Seal to this character.'
    )
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_charactor_status('Scarlet Seal')
        ]


class DoneDeal(ElementalBurstBase):
    name: Literal['Done Deal'] = 'Done Deal'
    desc: str = (
        'Deals 3 Pyro DMG, attaches Scarlet Seal and Brilliance to this '
        'character.'
    )
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_charactor_status('Scarlet Seal'),
            self.create_charactor_status('Brilliance'),
        ]


# Talents


class RightOfFinalInterpretation(SkillTalent):
    name: Literal['Right of Final Interpretation']
    desc: str = (
        'Combat Action: When your active character is Yanfei, equip this '
        'card. After Yanfei equips this card, immediately use Seal of '
        'Approval once. When Yanfei uses a Charged Attack with this card '
        'equipped: Deal +1 DMG to enemies with 6 or less HP.'
    )
    version: Literal['3.8'] = '3.8'
    charactor_name: Literal['Yanfei'] = 'Yanfei'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 1,
        any_dice_number = 2,
    )
    skill: SealOfApproval = SealOfApproval()

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        if not self.position.area == ObjectPositionType.CHARACTOR:
            # not equiped
            return value
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not self use normal attack
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charge attack
            return value
        if value.damage_from_element_reaction:  # pragma: no cover
            # damage from element reaction
            return value
        target = match.get_object(value.target_position)
        if target.hp > 6:
            # target hp > 6
            return value
        # increase damage
        assert mode == 'REAL'
        value.damage += 1
        return value


# charactor base


class Yanfei(CharactorBase):
    name: Literal['Yanfei']
    version: Literal['3.8'] = '3.8'
    desc: str = '''Wise Innocence: Yanfei'''
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        SealOfApproval | SignedEdict | DoneDeal
    ] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.CATALYST
    talent: RightOfFinalInterpretation | None = None

    def _init_skills(self) -> None:
        self.skills = [
            SealOfApproval(),
            SignedEdict(),
            DoneDeal()
        ]
